"""Collection of plots to evaluate a model, create overviews on data or forecasts."""
__author__ = "Lukas Leufen, Felix Kleinert"
__date__ = '2020-11-23'

import logging
import math
import os
import warnings
from typing import Dict, List, Tuple, Union
import itertools

import matplotlib
import matplotlib.pyplot as plt
import matplotlib.colors as colors
import numpy as np
import pandas as pd
import seaborn as sns
import xarray as xr
from matplotlib.backends.backend_pdf import PdfPages
from matplotlib.offsetbox import AnchoredText
import matplotlib.dates as mdates
from scipy.stats import mannwhitneyu
import datetime as dt

from mlair import helpers
from mlair.data_handler.iterator import DataCollection
from mlair.helpers import TimeTrackingWrapper
from mlair.helpers.helpers import relative_round
from mlair.plotting.abstract_plot_class import AbstractPlotClass
from mlair.helpers.statistics import mann_whitney_u_test, represent_p_values_as_asteriks


logging.getLogger('matplotlib').setLevel(logging.WARNING)


# import matplotlib
# matplotlib.use("TkAgg")
# import matplotlib.pyplot as plt


@TimeTrackingWrapper
class PlotMonthlySummary(AbstractPlotClass):  # pragma: no cover
    """
    Show a monthly summary over all stations for each lead time ("ahead") as box and whiskers plot.

    The plot is saved in data_path with name monthly_summary_box_plot.pdf and 500dpi resolution.

    .. image:: ../../../../../_source/_plots/monthly_summary_box_plot.png
        :width: 400

    :param stations: all stations to plot
    :param data_path: path, where the data is located
    :param name: full name of the local files with a % as placeholder for the station name
    :param target_var: display name of the target variable on plot's axis
    :param window_lead_time: lead time to plot, if window_lead_time is higher than the available lead time or not given
        the maximum lead time from data is used. (default None -> use maximum lead time from data).
    :param plot_folder: path to save the plot (default: current directory)
    :param target_var_unit: unit of target var for plot legend (default= ppb)

    """

    def __init__(self, stations: List, data_path: str, name: str, target_var: str, window_lead_time: int = None,
                 plot_folder: str = ".", target_var_unit: str = 'ppb', model_name="nn"):
        """Set attributes and create plot."""
        super().__init__(plot_folder, "monthly_summary_box_plot")
        self._data_path = data_path
        self._data_name = name
        self._model_name = model_name
        self._data = self._prepare_data(stations)
        self._window_lead_time = self._get_window_lead_time(window_lead_time)
        self._plot(target_var, target_var_unit)
        self._save()

    def _prepare_data(self, stations: List) -> xr.DataArray:
        """
        Pre.process data required to plot.

        For each station, load locally saved predictions, extract the CNN prediction and the observation and group them
        into monthly bins (no aggregation, only sorting them).

        :param stations: all stations to plot
        :return: The entire data set, flagged with the corresponding month.
        """
        forecasts = None
        for station in stations:
            logging.debug(f"... preprocess station {station}")
            file_name = os.path.join(self._data_path, self._data_name % station)
            data = xr.open_dataarray(file_name)

            data_nn = data.sel(type=self._model_name).squeeze()
            if len(data_nn.shape) > 1:
                data_nn = data_nn.assign_coords(ahead=[f"{days}d" for days in data_nn.coords["ahead"].values])
            else:
                data_nn.coords["ahead"].values = str(data_nn.coords["ahead"].values) + "d"

            data_obs = data.sel(type="obs", ahead=1).squeeze()
            data_obs.coords["ahead"] = "obs"

            data_concat = xr.concat([data_obs, data_nn], dim="ahead")
            data_concat = data_concat.drop_vars("type")

            new_index = data_concat.index.values.astype("datetime64[M]").astype(int) % 12 + 1
            data_concat = data_concat.assign_coords(index=new_index)
            data_concat = data_concat.clip(min=0)

            forecasts = xr.concat([forecasts, data_concat], 'index') if forecasts is not None else data_concat
        return forecasts

    def _get_window_lead_time(self, window_lead_time: int):
        """
        Extract the lead time from data and arguments.

        If window_lead_time is not given, extract this information from data itself by the number of ahead dimensions.
        If given, check if data supports the give length. If the number of ahead dimensions in data is lower than the
        given lead time, data's lead time is used.

        :param window_lead_time: lead time from arguments to validate
        :return: validated lead time, comes either from given argument or from data itself
        """
        ahead_steps = len(self._data.ahead)
        if window_lead_time is None:
            window_lead_time = ahead_steps
        return min(ahead_steps, window_lead_time)

    @staticmethod
    def _spell_out_chemical_concentrations(short_name: str):
        short2long = {'o3': 'ozone', 'no': 'nitrogen oxide', 'no2': 'nitrogen dioxide', 'nox': 'nitrogen dioxides'}
        return f"{short2long[short_name]} concentration"

    def _plot(self, target_var: str, target_var_unit: str):
        """
        Create a monthly grouped box plot over all stations but with separate boxes for each lead time step.

        :param target_var: display name of the target variable on plot's axis
        """
        data = self._data.to_dataset(name='values').to_dask_dataframe()
        logging.debug("... start plotting")
        color_palette = [colors.cnames["green"]] + sns.color_palette("Blues_d", self._window_lead_time).as_hex()
        ax = sns.boxplot(x='index', y='values', hue='ahead', data=data.compute(), whis=1.5, palette=color_palette,
                         flierprops={'marker': '.', 'markersize': 1}, showmeans=True,
                         meanprops={'markersize': 1, 'markeredgecolor': 'k'})
        ylabel = self._spell_out_chemical_concentrations(target_var)
        ax.set(xlabel='month', ylabel=f'{ylabel} (in {target_var_unit})')
        plt.tight_layout()


@TimeTrackingWrapper
class PlotConditionalQuantiles(AbstractPlotClass):  # pragma: no cover
    """
    Create cond.quantile plots as originally proposed by Murphy, Brown and Chen (1989) [But in log scale].

    Link to paper: https://journals.ametsoc.org/doi/pdf/10.1175/1520-0434%281989%29004%3C0485%3ADVOTF%3E2.0.CO%3B2

    .. image:: ../../../../../_source/_plots/conditional_quantiles_cali-ref_plot.png
        :width: 400

    .. image:: ../../../../../_source/_plots/conditional_quantiles_like-bas_plot.png
        :width: 400

    For each time step ahead a separate plot is created. If parameter plot_per_season is true, data is split by season
    and conditional quantiles are plotted for each season in addition.

    :param stations: all stations to plot
    :param data_pred_path: path to dir which contains the forecasts as .nc files
    :param plot_folder: path where the plots are stored
    :param plot_per_seasons: if `True' create cond. quantile plots for _seasons (DJF, MAM, JJA, SON) individually
    :param rolling_window: smoothing of quantiles (3 is used by Murphy et al.)
    :param model_name: name of the model prediction as stored in netCDF file (for example "nn")
    :param obs_name: name of observation as stored in netCDF file (for example "obs")
    :param kwargs: Some further arguments which are listed in self._opts
    """

    # ignore warnings if nans appear in quantile grouping
    warnings.filterwarnings("ignore", message="All-NaN slice encountered")
    # ignore warnings if mean is calculated on nans
    warnings.filterwarnings("ignore", message="Mean of empty slice")
    # ignore warnings for y tick = 0 on log scale (instead of 0.00001 or similar)
    warnings.filterwarnings("ignore", message="Attempted to set non-positive bottom ylim on a log-scaled axis.")

    def __init__(self, stations: List, data_pred_path: str, plot_folder: str = ".", plot_per_seasons=True,
                 rolling_window: int = 3, forecast_indicator: str = "nn", obs_indicator: str = "obs",
                 competitors=None, model_type_dim: str = "type", index_dim: str = "index", ahead_dim: str = "ahead",
                 competitor_path: str = None, sampling: str = "daily", model_name: str = "nn", **kwargs):
        """Initialise."""
        super().__init__(plot_folder, "conditional_quantiles")
        self._data_pred_path = data_pred_path
        self._stations = stations
        self._rolling_window = rolling_window
        self._forecast_indicator = forecast_indicator
        self.model_type_dim = model_type_dim
        self.index_dim = index_dim
        self.ahead_dim = ahead_dim
        self.iter_dim = "station"
        self.model_name = model_name
        self._obs_name = obs_indicator
        self._sampling = sampling
        self._opts = self._get_opts(kwargs)
        self._seasons = ['DJF', 'MAM', 'JJA', 'SON'] if plot_per_seasons is True else ""
        self.competitors = self._correct_persi_name(competitors or [])
        self.competitor_path = competitor_path or data_pred_path
        self._data = self._load_data()
        self._bins = self._get_bins_from_rage_of_data()
        self._plot()

    @staticmethod
    def _get_opts(kwargs):
        """Extract options from kwargs."""
        return {"q": kwargs.get("q", [.1, .25, .5, .75, .9]),
                "linetype": kwargs.get("linetype", [':', '-.', '--', '-.', ':']),
                "legend": kwargs.get("legend", ['.10th and .90th quantile', '.25th and .75th quantile',
                                                '.50th quantile', 'reference 1:1']),
                "data_unit": kwargs.get("data_unit", "ppb"), }

    def _load_data(self) -> xr.DataArray:
        """
        Load plot data.

        :return: plot data
        """
        logging.debug("... load data")
        data_collector = []
        for station in self._stations:
            file = os.path.join(self._data_pred_path, f"forecasts_{station}_test.nc")
            data_tmp = xr.open_dataarray(file)
            start = data_tmp.coords[self.index_dim].min().values
            end = data_tmp.coords[self.index_dim].max().values
            competitor = self.load_competitors(station, start, end)
            combined = self._combine_forecasts(data_tmp, competitor, dim=self.model_type_dim)
            sel = combined.sel({self.model_type_dim: [self._forecast_indicator, self._obs_name, *self.competitors]})
            data_collector.append(sel.assign_coords({self.iter_dim: station}))
        res = xr.concat(data_collector, dim=self.iter_dim).transpose(self.index_dim, self.model_type_dim,
                                                                     self.ahead_dim, self.iter_dim)
        return res

    def _combine_forecasts(self, forecast, competitor, dim=None):
        """
        Combine forecast and competitor if both are xarray. If competitor is None, this returns forecasts and vise
        versa.
        """
        if dim is None:
            dim = self.model_type_dim
        try:
            return xr.concat([forecast, competitor], dim=dim)
        except (TypeError, AttributeError):
            return forecast if competitor is None else competitor

    def load_competitors(self, station_name: str, start, end) -> xr.DataArray:
        """
        Load all requested and available competitors for a given station. Forecasts must be available in the competitor
        path like `<competitor_path>/<target_var>/forecasts_<station_name>_test.nc`. The naming style is equal for all
        forecasts of MLAir, so that forecasts of a different experiment can easily be copied into the competitor path
        without any change.

        :param station_name: station indicator to load competitors for

        :return: a single xarray with all competing forecasts
        """
        competing_predictions = []
        for competitor_name in self.competitors:
            try:
                prediction = self._create_competitor_forecast(station_name, competitor_name, start, end)
                competing_predictions.append(prediction)
            except (FileNotFoundError, KeyError):
                logging.debug(f"No competitor found for combination '{station_name}' and '{competitor_name}'.")
                continue
        return xr.concat(competing_predictions, self.model_type_dim) if len(competing_predictions) > 0 else None

    @staticmethod
    def create_full_time_dim(data, dim, sampling, start, end):
        """Ensure time dimension to be equidistant. Sometimes dates if missing values have been dropped."""
        start_data = data.coords[dim].values[0]
        freq = {"daily": "1D", "hourly": "1H"}.get(sampling)
        _ind = pd.date_range(start, end, freq=freq)  # two steps required to include all hours of end interval
        datetime_index = pd.DataFrame(index=pd.date_range(_ind.min(), _ind.max() + dt.timedelta(days=1),
                                                          closed="left", freq=freq))
        t = data.sel({dim: start_data}, drop=True)
        res = xr.DataArray(coords=[datetime_index.index, *[t.coords[c] for c in t.coords]], dims=[dim, *t.coords])
        res = res.transpose(*data.dims)
        if data.shape == res.shape:
            res.loc[data.coords] = data
        else:
            _d = data.sel({dim: slice(start, end)})
            res.loc[_d.coords] = _d
        return res

    def _create_competitor_forecast(self, station_name: str, competitor_name: str, start, end) -> xr.DataArray:
        """
        Load and format the competing forecast of a distinct model indicated by `competitor_name` for a distinct station
        indicated by `station_name`. The name of the competitor is set in the `type` axis as indicator. This method will
        raise either a `FileNotFoundError` or `KeyError` if no competitor could be found for the given station. Either
        there is no file provided in the expected path or no forecast for given `competitor_name` in the forecast file.
        Forecast is trimmed on interval start and end of test subset.

        :param station_name: name of the station to load data for
        :param competitor_name: name of the model
        :return: the forecast of the given competitor
        """
        path = os.path.join(self.competitor_path, competitor_name)
        file = os.path.join(path, f"forecasts_{station_name}_test.nc")
        with xr.open_dataarray(file) as da:
            data = da.load()
        if self._forecast_indicator in data.coords[self.model_type_dim]:
            forecast = data.sel({self.model_type_dim: [self._forecast_indicator]})
            forecast.coords[self.model_type_dim] = [competitor_name]
        else:
            forecast = data.sel({self.model_type_dim: [competitor_name]})
        # limit forecast to time range of test subset
        return self.create_full_time_dim(forecast, self.index_dim, self._sampling, start, end)

    @staticmethod
    def _correct_persi_name(competitors):
        return ["persi" if x == "Persistence" else x for x in competitors]

    def _segment_data(self, data: xr.DataArray, x_model: str) -> xr.DataArray:
        """
        Segment data into bins.

        :param data: data to segment
        :param x_model: name of x dimension

        :return: segmented data
        """
        logging.debug("... segment data")
        # combine index and station to multi index
        data = data.stack(z=[self.index_dim, self.iter_dim])
        # replace multi index by simple position index (order is not relevant anymore)
        data.coords['z'] = range(len(data.coords['z']))
        # segment data of x_model into bins
        data_sel = data.sel({self.model_type_dim: x_model})
        data.loc[{self.model_type_dim: x_model}] = data_sel.to_pandas().T.apply(pd.cut, bins=self._bins,
                                                                                labels=self._bins[1:]).T.values
        return data

    @staticmethod
    def _labels(plot_type: str, data_unit: str = "ppb") -> Tuple[str, str]:
        """
        Assign (x,y) labels to plots correctly, depending on like-base or cali-ref factorization.

        :param plot_type: type of plot, either `obs` or a model name
        :param data_unit: unit of data to add to labels (default ppb)

        :return: tuple with y and x labels
        """
        names = (f"forecasted concentration (in {data_unit})", f"observed concentration (in {data_unit})")
        if plot_type == "obs":
            return names
        else:
            return names[::-1]

    def _get_bins_from_rage_of_data(self) -> np.ndarray:
        """
        Get array of bins to use for quantiles.

        :return: range from 0 to data's maximum + 1 (rounded down)
        """
        return np.arange(0, math.ceil(self._data.max().max()) + 1, 1).astype(int)

    def _create_quantile_panel(self, data: xr.DataArray, x_model: str, y_model: str) -> xr.DataArray:
        """
        Calculate quantiles.

        :param data: data to calculate quantiles
        :param x_model: name of x dimension
        :param y_model: name of y dimension

        :return: quantile panel with binned data
        """
        logging.debug("... create quantile panel")
        # create empty xarray with dims: time steps ahead, quantiles, bin index (numbers create in previous step)
        quantile_panel = xr.DataArray(
            np.full([data.ahead.shape[0], len(self._opts["q"]), self._bins[1:].shape[0]], np.nan),
            coords=[data.ahead, self._opts["q"], self._bins[1:]], dims=[self.ahead_dim, 'quantiles', 'categories'])
        # ensure that the coordinates are in the right order
        quantile_panel = quantile_panel.transpose(self.ahead_dim, 'quantiles', 'categories')
        # calculate for each bin of the pred_name data the quantiles of the ref_name data
        for bin in self._bins[1:]:
            mask = (data.loc[x_model, ...] == bin)
            quantile_panel.loc[..., bin] = data.loc[y_model, ...].where(mask).quantile(self._opts["q"], dim=['z']).T
        return quantile_panel

    @staticmethod
    def add_affix(affix: str) -> str:
        """
        Add additional information to plot name with leading underscore or add empty string if affix is empty.

        :param affix: string to add

        :return: affix with leading underscore or empty string.
        """
        return f"_{affix}" if len(affix) > 0 else ""

    def _prepare_plots(self, data: xr.DataArray, x_model: str, y_model: str) -> Tuple[xr.DataArray, xr.DataArray]:
        """
        Get segmented data and quantile panel.

        :param data: plot data
        :param x_model: name of x dimension
        :param y_model: name of y dimension

        :return: segmented data and quantile panel
        """
        segmented_data = self._segment_data(data, x_model)
        quantile_panel = self._create_quantile_panel(segmented_data, x_model, y_model)
        return segmented_data, quantile_panel

    def _plot(self):
        """Start plotting routines: overall plot and seasonal (if enabled)."""
        logging.info(f"start plotting {self.__class__.__name__}, scheduled number of plots: "
                     f"{(len(self._seasons) + 1) * (len(self.competitors) + 1) * 2}")
        if len(self._seasons) > 0:
            self._plot_seasons()
        self._plot_all()

    def _plot_seasons(self):
        """Create seasonal plots."""
        for model in [self._forecast_indicator, *self.competitors]:
            for season in self._seasons:
                self._plot_base(data=self._data.where(self._data[f"{self.index_dim}.season"] == season),
                                x_model=model, y_model=self._obs_name, plot_name_affix="cali-ref",
                                season=season, model_name=model)
                self._plot_base(data=self._data.where(self._data[f"{self.index_dim}.season"] == season),
                                x_model=self._obs_name, y_model=model, plot_name_affix="like-base",
                                season=season, model_name=model)

    def _plot_all(self):
        """Plot overall conditional quantiles on full data."""
        for model in [self._forecast_indicator, *self.competitors]:
            self._plot_base(data=self._data, x_model=model, y_model=self._obs_name,
                            plot_name_affix="cali-ref", model_name=model)
            self._plot_base(data=self._data, x_model=self._obs_name, y_model=model,
                            plot_name_affix="like-base", model_name=model)

    @TimeTrackingWrapper
    def _plot_base(self, data: xr.DataArray, x_model: str, y_model: str, plot_name_affix: str, season: str = "",
                   model_name: str = ""):
        """
        Create conditional quantile plots.

        :param data: data which is used to create cond. quantile plot
        :param x_model: name of model on x axis (can also be obs)
        :param y_model: name of model on y axis (can also be obs)
        :param plot_name_affix: should be `cali-ref' or `like-base'
        :param season: List of _seasons to use
        """
        segmented_data, quantile_panel = self._prepare_plots(data, x_model, y_model)
        ylabel, xlabel = self._labels(x_model, self._opts["data_unit"])
        plot_name = f"{self.plot_name}{self.add_affix(season)}{self.add_affix(plot_name_affix)}" \
                    f"{self.add_affix(model_name)}.pdf"
        plot_path = os.path.join(os.path.abspath(self.plot_folder), plot_name)
        pdf_pages = matplotlib.backends.backend_pdf.PdfPages(plot_path)
        logging.debug(f"... plot path is {plot_path}")

        # create plot for each time step ahead
        y2_max = 0
        for iteration, d in enumerate(segmented_data.ahead):
            logging.debug(f"... plotting {d.values} time step(s) ahead")
            # plot smoothed lines with rolling mean
            smooth_data = quantile_panel.loc[d, ...].rolling(categories=self._rolling_window,
                                                             center=True).mean().to_pandas().T
            ax = smooth_data.plot(style=self._opts["linetype"], color='black', legend=False)
            ax2 = ax.twinx()
            # add reference line
            ax.plot([0, self._bins.max()], [0, self._bins.max()], color='k', label='reference 1:1', linewidth=.8)
            # add histogram of the segmented data (pred_name)
            handles, labels = ax.get_legend_handles_labels()
            segmented_data.loc[x_model, d, :].to_pandas().hist(bins=self._bins, ax=ax2, color='k', alpha=.3, grid=False,
                                                               rwidth=1)
            # add legend
            plt.legend(handles[:3] + [handles[-1]], self._opts["legend"], loc='upper left', fontsize='large')
            # adjust limits and set labels
            ax.set(xlim=(0, self._bins.max()), ylim=(0, self._bins.max()))
            ax.set_xlabel(xlabel, fontsize='x-large')
            ax.tick_params(axis='x', which='major', labelsize=15)
            ax.set_ylabel(ylabel, fontsize='x-large')
            ax.tick_params(axis='y', which='major', labelsize=15)
            ax2.yaxis.label.set_color('gray')
            ax2.tick_params(axis='y', colors='gray')
            ax2.yaxis.labelpad = -15
            ax2.set_yscale('log')
            if iteration == 0:
                y2_max = ax2.get_ylim()[1] + 100
            ax2.set(ylim=(0, y2_max * 10 ** 8), yticks=np.logspace(0, 4, 5))
            ax2.set_ylabel('              sample size', fontsize='x-large')
            ax2.tick_params(axis='y', which='major', labelsize=15)
            # set title and save current figure
            sampling_letter = {"daily": "D", "hourly": "H"}.get(self._sampling)
            model_name = self.model_name if model_name == self._forecast_indicator else model_name
            title = f"{model_name} ({sampling_letter}{d.values}{f', {season}' if len(season) > 0 else ''})"
            plt.title(title)
            pdf_pages.savefig()
        # close all open figures / plots
        pdf_pages.close()
        plt.close('all')


@TimeTrackingWrapper
class PlotClimatologicalSkillScore(AbstractPlotClass):  # pragma: no cover
    """
    Create plot of climatological skill score after Murphy (1988) as box plot over all stations.

    A forecast time step (called "ahead") is separately shown to highlight the differences for each prediction time
    step. Either each single term is plotted (score_only=False) or only the resulting scores CASE I to IV are displayed
    (score_only=True, default). Y-axis is adjusted following the data and not hard coded. The plot is saved under
    plot_folder path with name skill_score_clim_{extra_name_tag}{model_setup}.pdf and resolution of 500dpi.

    .. image:: ../../../../../_source/_plots/skill_score_clim_all_terms_CNN.png
        :width: 400

    .. image:: ../../../../../_source/_plots/skill_score_clim_CNN.png
        :width: 400

    :param data: dictionary with station names as keys and 2D xarrays as values, consist on axis ahead and terms.
    :param plot_folder: path to save the plot (default: current directory)
    :param score_only: if true plot only scores of CASE I to IV, otherwise plot all single terms (default True)
    :param extra_name_tag: additional tag that can be included in the plot name (default "")
    :param model_name: architecture type to specify plot name (default "")

    """

    def __init__(self, data: Dict, plot_folder: str = ".", score_only: bool = True, extra_name_tag: str = "",
                 model_name: str = ""):
        """Initialise."""
        super().__init__(plot_folder, f"skill_score_clim_{extra_name_tag}{model_name}")
        self._labels = None
        self._data = self._prepare_data(data, score_only)
        self._plot(score_only)
        self._save()

    def _prepare_data(self, data: Dict, score_only: bool) -> pd.DataFrame:
        """
        Shrink given data, if only scores are relevant.

        In any case, transform data to a plot friendly format. Also set plot labels depending on the lead time
        dimensions.

        :param data: dictionary with station names as keys and 2D xarrays as values
        :param score_only: if true only scores of CASE I to IV are relevant
        :return: pre-processed data set
        """
        data = helpers.dict_to_xarray(data, "station")
        self._labels = [str(i) + "d" for i in data.coords["ahead"].values]
        if score_only:
            data = data.loc[:, ["CASE I", "CASE II", "CASE III", "CASE IV"], :]
        return data.to_dataframe("data").reset_index(level=[0, 1, 2])

    def _label_add(self, score_only: bool):
        """
        Add the phrase "terms and " if score_only is disabled or empty string (if score_only=True).

        :param score_only: if false all terms are relevant, otherwise only CASE I to IV
        :return: additional label
        """
        return "" if score_only else "terms and "

    def _plot(self, score_only, xlim=5):
        """
        Plot climatological skill score.

        :param score_only: if true plot only scores of CASE I to IV, otherwise plot all single terms
        """
        fig, ax = plt.subplots()
        if not score_only:
            fig.set_size_inches(11.7, 8.27)
        sns.boxplot(x="terms", y="data", hue="ahead", data=self._data, ax=ax, whis=1.5, palette="Blues_d",
                    showmeans=True, meanprops={"markersize": 1, "markeredgecolor": "k"}, flierprops={"marker": "."})
        ax.axhline(y=0, color="grey", linewidth=.5)
        ax.set(ylabel=f"{self._label_add(score_only)}skill score", xlabel="", title="summary of all stations",
               ylim=self._lim())
        handles, _ = ax.get_legend_handles_labels()
        ax.legend(handles, self._labels)
        plt.tight_layout()

    def _lim(self) -> Tuple[float, float]:
        """
        Calculate axis limits from data (Can be used to set axis extend).

        Lower limit is the minimum of 0 and data's minimum (reduced by small subtrahend) and upper limit is data's
        maximum (increased by a small addend).

        :return:
        """
        limit = 5
        lower = np.max([-limit, np.min([0, helpers.float_round(self._data["data"].min() - 0.1, 2)])])
        upper = np.min([limit, helpers.float_round(self._data["data"].max() + 0.1, 2)])
        return lower, upper


@TimeTrackingWrapper
class PlotCompetitiveSkillScore(AbstractPlotClass):  # pragma: no cover
    """
    Create competitive skill score plot.

    Create this plot for the given model setup and the reference models ordinary least squared ("ols") and the
    persistence forecast ("persi") for all lead times ("ahead"). The plot is saved under plot_folder with the name
    skill_score_competitive_{model_setup}.pdf and resolution of 500dpi.

    .. image:: ../../../../../_source/_plots/skill_score_competitive.png
        :width: 400

    :param data: data frame with index=['cnn-persi', 'ols-persi', 'cnn-ols'] and columns "ahead" containing the pre-
        calculated comparisons for cnn, persistence and ols.
    :param plot_folder: path to save the plot (default: current directory)
    :param model_setup: architecture type (default "CNN")

    """

    def __init__(self, data: Dict[str, pd.DataFrame], plot_folder=".", model_setup="NN"):
        """Initialise."""
        super().__init__(plot_folder, f"skill_score_competitive_{model_setup}")
        self._model_setup = model_setup
        self._labels = None
        self._data = self._prepare_data(helpers.remove_items(data, "total"))
        default_plot_name = self.plot_name
        # draw full detail plot
        self.plot_name = default_plot_name + "_full_detail"
        self._plot()
        self._save()
        # draw also a vertical full detail version
        self.plot_name = default_plot_name + "_full_detail_vertical"
        self._plot_vertical()
        self._save()
        # draw default plot with only model comparison
        self.plot_name = default_plot_name
        self._plot(single_model_comparison=True)
        self._save()
        # draw also a vertical full detail version
        self.plot_name = default_plot_name + "_vertical"
        self._plot_vertical(single_model_comparison=True)
        self._save()

    def _prepare_data(self, data: pd.DataFrame) -> pd.DataFrame:
        """
        Reformat given data and create plot labels and introduce the dimensions stations and comparison.

        :param data: data frame with index=['cnn-persi', 'ols-persi', 'cnn-ols'] and columns "ahead" containing the pre-
            calculated comparisons for cnn, persistence and ols.
        :return: processed data
        """
        data = pd.concat(data, axis=0)
        data = xr.DataArray(data, dims=["stations", "ahead"]).unstack("stations")
        data = data.rename({"stations_level_0": "stations", "stations_level_1": "comparison"})
        data = data.to_dataframe("data").unstack(level=1).swaplevel()
        data.columns = data.columns.levels[1]
        self._labels = [str(i) + "d" for i in data.index.levels[1].values]
        data = data.stack(level=0).reset_index(level=2, drop=True).reset_index(name="data")
        return data.astype({"comparison": str, "ahead": int, "data": float})

    def _plot(self, single_model_comparison=False):
        """Plot skill scores of the comparisons."""
        data = self._filter_comparisons(self._data) if single_model_comparison is True else self._data
        max_label_size = len(max(np.unique(data.comparison).tolist(), key=len))
        size = max([len(np.unique(data.comparison)), 6])
        fig, ax = plt.subplots(figsize=(size, 5 * max(0.8, max_label_size/20)))
        order = self._create_pseudo_order(data)
        sns.boxplot(x="comparison", y="data", hue="ahead", data=data, whis=1.5, ax=ax, palette="Blues_d",
                    showmeans=True, meanprops={"markersize": 3, "markeredgecolor": "k"}, flierprops={"marker": "."},
                    order=order)
        ax.axhline(y=0, color="grey", linewidth=.5)
        ax.set(ylabel="skill score", xlabel="competing models", title="summary of all stations", ylim=self._lim(data))
        handles, _ = ax.get_legend_handles_labels()
        plt.xticks(rotation=90)
        ax.legend(handles, self._labels)
        plt.tight_layout()

    def _plot_vertical(self, single_model_comparison=False):
        """Plot skill scores of the comparisons, but vertically aligned."""
        data = self._filter_comparisons(self._data) if single_model_comparison is True else self._data
        max_label_size = len(max(np.unique(data.comparison).tolist(), key=len))
        size = max([len(np.unique(data.comparison)), 6])
        fig, ax = plt.subplots(figsize=(5 * max(0.8, max_label_size/20), size))
        order = self._create_pseudo_order(data)
        sns.boxplot(y="comparison", x="data", hue="ahead", data=data, whis=1.5, ax=ax, palette="Blues_d",
                    showmeans=True, meanprops={"markersize": 3, "markeredgecolor": "k"}, flierprops={"marker": "."},
                    order=order)
        ax.axvline(x=0, color="grey", linewidth=.5)
        ax.set(xlabel="skill score", ylabel="competing models", title="summary of all stations", xlim=self._lim(data))
        handles, _ = ax.get_legend_handles_labels()
        ax.legend(handles, self._labels)
        plt.tight_layout()

    def _create_pseudo_order(self, data):
        """Provide first predefined elements and append all remaining."""
        first_elements = [f"{self._model_setup} - persi", "ols - persi", f"{self._model_setup} - ols"]
        first_elements = list(filter(lambda x: x in data.comparison.tolist(), first_elements))
        uniq, index = np.unique(first_elements + data.comparison.unique().tolist(), return_index=True)
        return uniq[index.argsort()]

    def _filter_comparisons(self, data):
        filtered_headers = list(filter(lambda x: f"{self._model_setup} - " in x, data.comparison.unique()))
        return data[data.comparison.isin(filtered_headers)]

    @staticmethod
    def _lim(data) -> Tuple[float, float]:
        """
        Calculate axis limits from data (Can be used to set axis extend).

        Lower limit is the minimum of 0 and data's minimum (reduced by small subtrahend) and upper limit is data's
        maximum (increased by a small addend).

        :return:
        """
        limit = 5
        lower = np.max([-limit, np.min([0, helpers.float_round(data.min()[2], 2) - 0.1])])
        upper = np.min([limit, helpers.float_round(data.max()[2], 2) + 0.1])
        return lower, upper


@TimeTrackingWrapper
class PlotFeatureImportanceSkillScore(AbstractPlotClass):  # pragma: no cover
    """
    Create plot of feature importance analysis.

    By passing a list `separate_vars` containing variable names, a second plot is created showing the `separate_vars`
    and the remaining variables side by side with different scaling.

    .. image:: ../../../../../_source/_plots/skill_score_bootstrap.png
        :width: 400

    .. image:: ../../../../../_source/_plots/skill_score_bootstrap_separated.png
        :width: 400

    """

    def __init__(self, data: Dict, plot_folder: str = ".", separate_vars: List = None, sampling: str = "daily",
                 ahead_dim: str = "ahead", bootstrap_type: str = None, bootstrap_method: str = None,
                 boot_dim: str = "boots", model_name: str = "NN", branch_names: list = None, ylim: tuple = None):
        """
        Set attributes and create plot.

        :param data: dictionary with station names as keys and 2D xarrays as values, consist on axis ahead and terms.
        :param plot_folder: path to save the plot (default: current directory)
        :param separate_vars: variables to plot separated (default: ['o3'])
        :param sampling: type of sampling rate, should be either hourly or daily (default: "daily")
        :param ahead_dim: name of the ahead dimensions (default: "ahead")
        :param bootstrap_annotation: additional information to use in the file name (default: None)
        :param model_name: architecture type to specify plot name (default "NN")
        """
        annotation = ["_".join([s for s in ["", bootstrap_type, bootstrap_method] if s is not None])][0]
        super().__init__(plot_folder, f"feature_importance_{model_name}{annotation}")
        if separate_vars is None:
            separate_vars = ['o3']
        self._labels = None
        self._x_name = "boot_var"
        self._ahead_dim = ahead_dim
        self._boot_dim = boot_dim
        self._boot_type = self._set_bootstrap_type(bootstrap_type)
        self._boot_method = self._set_bootstrap_method(bootstrap_method)
        self._number_of_bootstraps = 0
        self._branches_names = branch_names
        self._ylim = ylim

        self._data = self._prepare_data(data, sampling)
        self._set_title(model_name)
        if "branch" in self._data.columns:
            plot_name = self.plot_name
            for branch in self._data["branch"].unique():
                self._set_title(model_name, branch, len(self._data["branch"].unique()))
                self.plot_name = f"{plot_name}_{branch}"
                try:
                    self._plot(branch=branch)
                    self._save()
                except ValueError as e:
                    logging.info(f"Did not plot {self.plot_name} because of {e}")
                if len(set(separate_vars).intersection(self._data[self._x_name].unique())) > 0:
                    self.plot_name += '_separated'
                    try:
                        self._plot(branch=branch, separate_vars=separate_vars)
                        self._save(bbox_inches='tight')
                    except ValueError as e:
                        logging.info(f"Did not plot {self.plot_name} because of {e}")
        else:
            try:
                self._plot()
                self._save()
            except ValueError as e:
                logging.info(f"Did not plot {self.plot_name} because of {e}")
            if len(set(separate_vars).intersection(self._data[self._x_name].unique())) > 0:
                self.plot_name += '_separated'
                try:
                    self._plot(separate_vars=separate_vars)
                    self._save(bbox_inches='tight')
                except ValueError as e:
                    logging.info(f"Did not plot {self.plot_name} because of {e}")

    @staticmethod
    def _set_bootstrap_type(boot_type):
        return {"singleinput": "single input"}.get(boot_type, boot_type)

    def _set_title(self, model_name, branch=None, n_branches=None):
        title_d = {"single input": "Single Inputs", "branch": "Input Branches", "variable": "Variables"}
        base_title = f"{model_name}\nImportance of {title_d[self._boot_type]}"

        additional = []
        if branch is not None:
            try:
                assert n_branches == len(self._branches_names)
                branch_name = self._branches_names[int(branch)]
            except (IndexError, TypeError, ValueError, AssertionError):
                branch_name = branch
            additional.append(branch_name)
        if self._number_of_bootstraps > 1:
            additional.append(f"n={self._number_of_bootstraps}")
        additional_title = ", ".join(additional)
        if len(additional_title) > 0:
            additional_title = f" ({additional_title})"
        self._title = base_title + additional_title

    @staticmethod
    def _set_bootstrap_method(boot_method):
        return {"zero_mean": "zero mean", "shuffle": "shuffled"}.get(boot_method, boot_method)

    def _prepare_data(self, data: Dict, sampling: str) -> pd.DataFrame:
        """
        Shrink given data, if only scores are relevant.

        In any case, transform data to a plot friendly format. Also set plot labels depending on the lead time
        dimensions.

        :param data: dictionary with station names as keys and 2D xarrays as values
        :return: pre-processed data set
        """
        station_dim = "station"
        data = helpers.dict_to_xarray(data, station_dim).sortby(self._x_name)
        data = data.transpose(station_dim, self._ahead_dim, self._boot_dim, self._x_name)
        if self._boot_type == "single input":
            number_tags = self._get_number_tag(data.coords[self._x_name].values, split_by='_')
            new_boot_coords = self._return_vars_without_number_tag(data.coords[self._x_name].values, split_by='_',
                                                                   keep=1, as_unique=True)
            try:
                values = data.values.reshape((*data.shape[:3], len(number_tags), len(new_boot_coords)))
                data = xr.DataArray(values, coords={station_dim: data.coords[station_dim], self._x_name: new_boot_coords,
                                                    "branch": number_tags, self._ahead_dim: data.coords[self._ahead_dim],
                                                    self._boot_dim: data.coords[self._boot_dim]},
                                    dims=[station_dim, self._ahead_dim, self._boot_dim, "branch", self._x_name])
            except ValueError:
                data_coll = []
                for nr in number_tags:
                    filtered_coords = list(filter(lambda x: nr in x.split("_")[0], data.coords[self._x_name].values))
                    new_boot_coords = self._return_vars_without_number_tag(filtered_coords, split_by='_', keep=1,
                                                                           as_unique=True)
                    sel_data = data.sel({self._x_name: filtered_coords})
                    values = sel_data.values.reshape((*data.shape[:3], 1, len(new_boot_coords)))
                    sel_data = xr.DataArray(values, coords={station_dim: data.coords[station_dim], self._x_name: new_boot_coords,
                                                    "branch": [nr], self._ahead_dim: data.coords[self._ahead_dim],
                                                    self._boot_dim: data.coords[self._boot_dim]},
                                        dims=[station_dim, self._ahead_dim, self._boot_dim, "branch", self._x_name])
                    data_coll.append(sel_data)
                data = xr.concat(data_coll, "branch")
        else:
            try:
                new_boot_coords = self._return_vars_without_number_tag(data.coords[self._x_name].values, split_by='_',
                                                                       keep=1)
                data = data.assign_coords({self._x_name: new_boot_coords})
            except NotImplementedError:
                pass
        _, sampling_letter = self._get_sampling(sampling, 1)
        sampling_letter = {"d": "D", "h": "H"}.get(sampling_letter, sampling_letter)
        self._labels = [sampling_letter + str(i) for i in data.coords[self._ahead_dim].values]
        if station_dim not in data.dims:
            data = data.expand_dims(station_dim)
        self._number_of_bootstraps = np.unique(data.coords[self._boot_dim].values).shape[0]
        return data.to_dataframe("data").reset_index(level=np.arange(len(data.dims)).tolist()).dropna()

    def _return_vars_without_number_tag(self, values, split_by, keep, as_unique=False):
        arr = np.array([v.split(split_by) for v in values])
        num = arr[:, 0]
        if arr.shape[keep] == 1:  # keep dim has only length 1, no number tags required
            return num
        new_val = arr[:, keep]
        if self._all_values_are_equal(num, axis=0):
            return new_val
        elif as_unique is True:
            return np.unique(new_val)
        else:
            raise NotImplementedError

    @staticmethod
    def _get_number_tag(values, split_by):
        arr = np.array([v.split(split_by) for v in values])
        num = arr[:, 0]
        return np.unique(num).tolist()

    @staticmethod
    def _all_values_are_equal(arr, axis=0):
        if np.all(arr == arr[0], axis=axis):
            return True
        else:
            return False

    def _label_add(self, score_only: bool):
        """
        Add the phrase "terms and " if score_only is disabled or empty string (if score_only=True).

        :param score_only: if false all terms are relevant, otherwise only CASE I to IV
        :return: additional label
        """
        return "" if score_only else "terms and "

    def _plot(self, branch=None, separate_vars=None):
        """Plot climatological skill score."""
        if separate_vars is None:
            self._plot_all_variables(branch)
        else:
            self._plot_selected_variables(separate_vars, branch)

    def _plot_selected_variables(self, separate_vars: List, branch=None):
        data = self._data if branch is None else self._data[self._data["branch"] == str(branch)]
        self.raise_error_if_vars_do_not_exist(data, separate_vars, self._x_name, name="separate_vars")
        all_variables = self._get_unique_values_from_column_of_df(data, self._x_name)
        remaining_vars = helpers.remove_items(all_variables, separate_vars)
        self.raise_error_if_vars_do_not_exist(data, remaining_vars, self._x_name, name="remaining_vars")
        data_first = self._select_data(df=data, variables=separate_vars, column_name=self._x_name)
        data_second = self._select_data(df=data, variables=remaining_vars, column_name=self._x_name)

        fig, ax = plt.subplots(nrows=1, ncols=2, gridspec_kw={'width_ratios': [len(separate_vars),
                                                                               len(remaining_vars)]},
                               figsize=(len(remaining_vars),len(remaining_vars)/2.))
        if len(separate_vars) > 1:
            first_box_width = .8
        else:
            first_box_width = .8

        sns.boxplot(x=self._x_name, y="data", hue=self._ahead_dim, data=data_first, ax=ax[0], whis=1.5,
                    palette="Blues_d", showmeans=True, meanprops={"markersize": 1, "markeredgecolor": "k"},
                    showfliers=False, width=first_box_width)
        ax[0].set(ylabel=f"skill score", xlabel="")
        if self._ylim is not None:
            _ylim = self._ylim if isinstance(self._ylim, tuple) else self._ylim[0]
            ax[0].set(ylim=_ylim)

        sns.boxplot(x=self._x_name, y="data", hue=self._ahead_dim, data=data_second, ax=ax[1], whis=1.5,
                    palette="Blues_d", showmeans=True, meanprops={"markersize": 1, "markeredgecolor": "k"},
                    showfliers=False)
        ax[1].set(ylabel="", xlabel="")
        ax[1].yaxis.tick_right()
        if self._ylim is not None and isinstance(self._ylim, list):
            _ylim = self._ylim[1]
            ax[1].set(ylim=_ylim)

        handles, _ = ax[1].get_legend_handles_labels()
        for sax in ax:
            matplotlib.pyplot.sca(sax)
            sax.axhline(y=0, color="grey", linewidth=.5)
            plt.xticks(rotation=45, ha='right')
            sax.legend_.remove()

        # fig.legend(handles, self._labels, loc='upper center', ncol=len(handles) + 1, )
        ax[1].legend(handles, self._labels, loc='lower center', ncol=len(handles) + 1, fontsize="medium")

        def align_yaxis(ax1, ax2):
            """
            Align zeros of the two axes, zooming them out by same ratio

            This function is copy pasted from https://stackoverflow.com/a/41259922
            """
            axes = (ax1, ax2)
            extrema = [ax.get_ylim() for ax in axes]
            tops = [extr[1] / (extr[1] - extr[0]) for extr in extrema]
            # Ensure that plots (intervals) are ordered bottom to top:
            if tops[0] > tops[1]:
                axes, extrema, tops = [list(reversed(l)) for l in (axes, extrema, tops)]

            # How much would the plot overflow if we kept current zoom levels?
            tot_span = tops[1] + 1 - tops[0]

            b_new_t = extrema[0][0] + tot_span * (extrema[0][1] - extrema[0][0])
            t_new_b = extrema[1][1] - tot_span * (extrema[1][1] - extrema[1][0])
            axes[0].set_ylim(extrema[0][0], b_new_t)
            axes[1].set_ylim(t_new_b, extrema[1][1])

        align_yaxis(ax[0], ax[1])
        align_yaxis(ax[0], ax[1])
        plt.subplots_adjust(right=0.8)
        plt.title(self._title)

    @staticmethod
    def _select_data(df: pd.DataFrame, variables: List[str], column_name: str) -> pd.DataFrame:
        selected_data = None
        for i, variable in enumerate(variables):
            if i == 0:
                selected_data = df.loc[df[column_name] == variable]
            else:
                tmp_var = df.loc[df[column_name] == variable]
                selected_data = pd.concat([selected_data, tmp_var], axis=0)
        return selected_data

    def raise_error_if_vars_do_not_exist(self, data, vars, column_name, name="separate_vars"):
        if len(vars) == 0:
            msg = f"No variables are given for `{name}' to check in `self.data' "
            raise ValueError(msg)
        if not self._variables_exist_in_df(df=data, variables=vars, column_name=column_name):
            msg = f"At least one entry of `{name}' does not exist in `self.data' "
            raise ValueError(msg)

    @staticmethod
    def _get_unique_values_from_column_of_df(df: pd.DataFrame, column_name: str) -> List:
        return list(df[column_name].unique())

    def _variables_exist_in_df(self, df: pd.DataFrame, variables: List[str], column_name: str):
        vars_in_df = set(self._get_unique_values_from_column_of_df(df, column_name))
        return set(variables).issubset(vars_in_df)

    def _plot_all_variables(self, branch=None):
        """

        """
        plot_data = self._data if branch is None else self._data[self._data["branch"] == str(branch)]
        if self._boot_type == "branch":
            fig, ax = plt.subplots(figsize=(0.5 + 2 / len(plot_data[self._x_name].unique()) + len(plot_data[self._x_name].unique()),4))
            sns.boxplot(x=self._x_name, y="data", hue=self._ahead_dim, data=plot_data, ax=ax, whis=1.,
                        palette="Blues_d", showmeans=True, meanprops={"markersize": 1, "markeredgecolor": "k"},
                        showfliers=False, width=0.8)
        else:
            fig, ax = plt.subplots()
            sns.boxplot(x=self._x_name, y="data", hue=self._ahead_dim, data=plot_data, ax=ax, whis=1.5, palette="Blues_d",
                        showmeans=True, meanprops={"markersize": 1, "markeredgecolor": "k"}, showfliers=False)
        ax.axhline(y=0, color="grey", linewidth=.5)

        if self._ylim is not None:
            if isinstance(self._ylim, tuple):
                _ylim = self._ylim
            else:
                _ylim = (min(self._ylim[0][0], self._ylim[1][0]), max(self._ylim[0][1], self._ylim[1][1]))
            ax.set(ylim=_ylim)

        if self._boot_type == "branch":
            plt.xticks()
        else:
            plt.xticks(rotation=45)
        ax.set(ylabel=f"skill score", xlabel="", title=self._title)
        handles, _ = ax.get_legend_handles_labels()
        ax.legend(handles, self._labels)
        plt.tight_layout()


@TimeTrackingWrapper
class PlotTimeSeries:  # pragma: no cover
    """
    Create time series plot.

    Currently, plots are under development and not well designed for any use in public.
    """

    def __init__(self, stations: List, data_path: str, name: str, window_lead_time: int = None, plot_folder: str = ".",
                 sampling="daily", model_name="nn", obs_name="obs", ahead_dim="ahead"):
        """Initialise."""
        self._data_path = data_path
        self._data_name = name
        self._stations = stations
        self._model_name = model_name
        self._obs_name = obs_name
        self._ahead_dim = ahead_dim
        self._window_lead_time = self._get_window_lead_time(window_lead_time)
        self._sampling = self._get_sampling(sampling)
        self._plot(plot_folder)

    @staticmethod
    def _get_sampling(sampling):
        if sampling == "daily":
            return "D"
        elif sampling == "hourly":
            return "h"

    def _get_window_lead_time(self, window_lead_time: int):
        """
        Extract the lead time from data and arguments.

        If window_lead_time is not given, extract this information from data itself by the number of ahead dimensions.
        If given, check if data supports the give length. If the number of ahead dimensions in data is lower than the
        given lead time, data's lead time is used.

        :param window_lead_time: lead time from arguments to validate
        :return: validated lead time, comes either from given argument or from data itself
        """
        ahead_steps = len(self._load_data(self._stations[0]).coords[self._ahead_dim])
        if window_lead_time is None:
            window_lead_time = ahead_steps
        return min(ahead_steps, window_lead_time)

    def _load_data(self, station):
        logging.debug(f"... preprocess station {station}")
        file_name = os.path.join(self._data_path, self._data_name % station)
        data = xr.open_dataarray(file_name)
        return data.sel(type=[self._model_name, self._obs_name])

    def _plot(self, plot_folder):
        pdf_pages = self._create_pdf_pages(plot_folder)
        for pos, station in enumerate(self._stations):
            data = self._load_data(station)
            start, end = self._get_time_range(data)
            fig, axes, factor = self._create_subplots(start, end)
            nan_list = []
            for i_year in range(end - start + 1):
                data_year = data.sel(index=f"{start + i_year}")
                for i_half_of_year in range(factor):
                    pos = factor * i_year + i_half_of_year
                    try:
                        plot_data = self._create_plot_data(data_year, factor, i_half_of_year)
                        self._plot_obs(axes[pos], plot_data)
                        self._plot_ahead(axes[pos], plot_data)
                        if np.isnan(plot_data.values).all():
                            nan_list.append(pos)
                    except Exception:
                        nan_list.append(pos)
            self._clean_up_axes(nan_list, axes, fig)
            self._save_page(station, pdf_pages)
        pdf_pages.close()
        plt.close('all')

    @staticmethod
    def _clean_up_axes(nan_list, axes, fig):
        for i in reversed(nan_list):
            fig.delaxes(axes[i])

    @staticmethod
    def _save_page(station, pdf_pages):
        plt.suptitle(station)
        plt.legend()
        plt.tight_layout()
        pdf_pages.savefig(dpi=500)

    @staticmethod
    def _create_plot_data(data, factor, running_index):
        if factor > 1:
            if running_index == 0:
                data = data.where(data["index.month"] < 7)
            else:
                data = data.where(data["index.month"] >= 7)
        return data

    def _create_subplots(self, start, end):
        factor = 1
        if self._sampling == "h":
            factor = 2
        f, ax = plt.subplots((end - start + 1) * factor, sharey=True, figsize=(50, 30), squeeze=False)
        return f, ax[:, 0], factor

    def _plot_ahead(self, ax, data):
        color = sns.color_palette("Blues_d", self._window_lead_time).as_hex()
        for ahead in data.coords[self._ahead_dim].values:
            plot_data = data.sel({"type": self._model_name, self._ahead_dim: ahead}).drop(["type", self._ahead_dim]).squeeze().shift(index=ahead)
            sampling_letter = {"d": "D", "h": "H"}.get(self._sampling, self._sampling)
            label = f"{sampling_letter}{ahead}"
            ax.plot(plot_data, color=color[ahead - 1], label=label)

    def _plot_obs(self, ax, data):
        ahead = 1
        obs_data = data.sel(type="obs", ahead=ahead).shift(index=ahead)
        ax.plot(obs_data, color=colors.cnames["green"], label="obs")

    @staticmethod
    def _get_time_range(data):
        def f(x, f_x):
            return pd.to_datetime(f_x(x.index.values)).year

        return f(data, min), f(data, max)

    @staticmethod
    def _create_pdf_pages(plot_folder: str):
        """
        Store plot locally.

        :param plot_folder: path to save the plot
        """
        plot_name = os.path.join(os.path.abspath(plot_folder), 'timeseries_plot.pdf')
        logging.debug(f"... save plot to {plot_name}")
        return matplotlib.backends.backend_pdf.PdfPages(plot_name)


@TimeTrackingWrapper
class PlotSeparationOfScales(AbstractPlotClass):  # pragma: no cover

    def __init__(self, collection: DataCollection, plot_folder: str = ".", time_dim="datetime", window_dim="window",
                 filter_dim="filter", target_dim="variables"):
        """Initialise."""
        # create standard Gantt plot for all stations (currently in single pdf file with single page)
        plot_folder = os.path.join(plot_folder, "separation_of_scales")
        super().__init__(plot_folder, "separation_of_scales")
        self.time_dim = time_dim
        self.window_dim = window_dim
        self.filter_dim = filter_dim
        self.target_dim = target_dim
        self._plot(collection)

    def _plot(self, collection: DataCollection):
        orig_plot_name = self.plot_name
        for dh in collection:
            data = dh.get_X(as_numpy=False)[0]
            station = dh.id_class.station[0]
            data = data.sel(Stations=station)
            data.plot(x=self.time_dim, y=self.window_dim, col=self.filter_dim, row=self.target_dim, robust=True)
            self.plot_name = f"{orig_plot_name}_{station}"
            self._save()


@TimeTrackingWrapper
class PlotSampleUncertaintyFromBootstrap(AbstractPlotClass):  # pragma: no cover

    def __init__(self, data: xr.DataArray, plot_folder: str = ".", model_type_dim: str = "type",
                 error_measure: str = "mse", error_unit: str = None, dim_name_boots: str = 'boots',
                 block_length: str = None, model_name: str = "NN", model_indicator: str = "nn",
                 ahead_dim: str = "ahead", sampling: Union[str, Tuple[str]] = "", season_annotation: str = None,
                 apply_root: bool = True, plot_name="sample_uncertainty_from_bootstrap"):
        super().__init__(plot_folder, plot_name)
        self.default_plot_name = self.plot_name
        self.model_type_dim = model_type_dim
        self.ahead_dim = ahead_dim
        self.error_measure = error_measure
        self.dim_name_boots = dim_name_boots
        self.error_unit = error_unit
        self.block_length = block_length
        self.model_name = model_name
        _season = season_annotation or ""
        self.sampling = {"daily": "d", "hourly": "H"}.get(sampling[1] if isinstance(sampling, tuple) else sampling, "")
        data = self.rename_model_indicator(data, model_name, model_indicator)
        self.prepare_data(data)

        # create all combinations to plot (h/v, utest/notest, single/multi)
        variants = list(itertools.product(*[["v", "h"], [True, False], ["single", "multi", "panel"]]))

        # plot raw metric (mse)
        for orientation, utest, agg_type in variants:
            self._plot(orientation=orientation, apply_u_test=utest, agg_type=agg_type, season=_season)
            self._plot_kde(agg_type=agg_type, season=_season)

        if apply_root is True:
            # plot root of metric (rmse)
            self._apply_root()
            for orientation, utest, agg_type in variants:
                self._plot(orientation=orientation, apply_u_test=utest, agg_type=agg_type, tag="_sqrt", season=_season)
                self._plot_kde(agg_type=agg_type, tag="_sqrt", season=_season)

        self._data_table = None
        self._n_boots = None
        self._factor = None

    @property
    def get_asteriks_from_mann_whitney_u_result(self):
        return represent_p_values_as_asteriks(mann_whitney_u_test(data=self._data_table,
                                                                  reference_col_name=self.model_name,
                                                                  axis=0, alternative="two-sided").iloc[-1])

    def rename_model_indicator(self, data, model_name, model_indicator):
        data.coords[self.model_type_dim] = [{model_indicator: model_name}.get(n, n)
                                            for n in data.coords[self.model_type_dim].values]
        return data

    def prepare_data(self, data: xr.DataArray):
        data_table = data.to_dataframe(self.model_type_dim).unstack()
        factor = len(data.coords[self.ahead_dim]) if self.ahead_dim in data.dims else 1
        self._data_table = data_table[data_table.mean().sort_values().index].droplevel(0, axis=1)
        self._n_boots = int(self._data_table.shape[0] / factor)
        self._factor = factor

    def _apply_root(self):
        self._data_table = np.sqrt(self._data_table)
        self.error_measure = f"root {self.error_measure}"
        self.error_unit = self.error_unit.replace("$^2$", "")

    def _plot_kde(self, agg_type="single", tag="", season=""):
        self.plot_name = self.default_plot_name + "_kde" + "_" + agg_type + tag + {"": ""}.get(season, f"_{season}")
        data_table = self._data_table
        if agg_type == "multi":
            return  # nothing to do
        if self.ahead_dim not in data_table.index.names and agg_type == "panel":
            return  # nothing to do
        n_boots = self._n_boots
        error_label = self.error_measure if self.error_unit is None else f"{self.error_measure} (in {self.error_unit})"
        sampling_letter = {"d": "D", "h": "H"}.get(self.sampling, self.sampling)
        if agg_type == "single":
            fig, ax = plt.subplots()
            if self.ahead_dim in data_table.index.names:
                data_table = data_table.groupby(level=0).mean()
            sns.kdeplot(data=data_table, ax=ax)
            ylims = list(ax.get_ylim())
            ax.set_ylim([ylims[0], ylims[1]*1.025])
            ax.set_xlabel(error_label)

            text = f"n={n_boots}"
            if self.block_length is not None:
                text = f"{self.block_length}, {text}"
            if len(season) > 0:
                text = f"{season}, {text}"
            loc = "upper left"
            text_box = AnchoredText(text, frameon=True, loc=loc, pad=0.5, bbox_to_anchor=(0., 1.0),
                                    bbox_transform=ax.transAxes)
            plt.setp(text_box.patch, edgecolor='k', facecolor='w')
            ax.add_artist(text_box)
            ax.get_legend().set_title(None)
            plt.tight_layout()
        else:
            g = sns.FacetGrid(data_table.stack(self.model_type_dim).reset_index(), **{"col": self.ahead_dim},
                              hue=self.model_type_dim, legend_out=True)
            g.map(sns.kdeplot, 0)
            g.add_legend(title="")
            fig = plt.gcf()
            _labels = [sampling_letter + str(i) for i in data_table.index.levels[1].values]
            for axi, title in zip(g.axes.flatten(), _labels):
                axi.set_title(title)
            for axi in g.axes.flatten():
                axi.set_xlabel(None)
            fig.supxlabel(error_label)
            text = f"n={n_boots}"
            if self.block_length is not None:
                text = f"{self.block_length}, {text}"
            if len(season) > 0:
                text = f"{season}, {text}"
            loc = "upper right"
            text_box = AnchoredText(text, frameon=True, loc=loc)
            plt.setp(text_box.patch, edgecolor='k', facecolor='w')
            g.axes.flatten()[0].add_artist(text_box)
        self._save()
        plt.close("all")

    def _plot(self, orientation: str = "v", apply_u_test: bool = False, agg_type="single", tag="", season=""):
        self.plot_name = self.default_plot_name + {"v": "_vertical", "h": "_horizontal"}[orientation] + \
                         {True: "_u_test", False: ""}[apply_u_test] + "_" + agg_type + tag + \
                         {"": ""}.get(season, f"_{season}")
        if apply_u_test is True and agg_type == "multi":
            return  # not implemented
        data_table = self._data_table
        if self.ahead_dim not in data_table.index.names and agg_type in ["multi", "panel"]:
            return  # nothing to do
        if apply_u_test is True and agg_type == "panel":
            return  # nothing to do
        n_boots = self._n_boots
        size = len(np.unique(data_table.columns))
        asteriks = self.get_asteriks_from_mann_whitney_u_result if apply_u_test is True else None
        color_palette = sns.color_palette("Blues_d", self._factor).as_hex()
        sampling_letter = {"d": "D", "h": "H"}.get(self.sampling, self.sampling)
        if orientation == "v":
            figsize, width = (size, 5), 0.4
        elif orientation == "h":
            if agg_type == "multi":
                size *= np.sqrt(len(data_table.index.unique(self.ahead_dim)))
                size = max(size, 8)
            figsize, width = (7, (1+.5*size)), 0.65
        else:
            raise ValueError(f"orientation must be `v' or `h' but is: {orientation}")
        fig, ax = plt.subplots(figsize=figsize)
        if agg_type == "single":
            if self.ahead_dim in data_table.index.names:
                data_table = data_table.groupby(level=0).mean()
            sns.boxplot(data=data_table, ax=ax, whis=1.5, color="white",
                        showmeans=True, meanprops={"markersize": 6, "markeredgecolor": "k"},
                        flierprops={"marker": "o", "markerfacecolor": "black", "markeredgecolor": "none", "markersize": 3},
                        boxprops={'facecolor': 'none', 'edgecolor': 'k'}, width=width, orient=orientation)
        elif agg_type == "multi":
            xy = {"x": self.model_type_dim, "y": 0} if orientation == "v" else {"x": 0, "y": self.model_type_dim}
            sns.boxplot(data=data_table.stack(self.model_type_dim).reset_index(), ax=ax, whis=1.5, palette=color_palette,
                        showmeans=True, meanprops={"markersize": 6, "markeredgecolor": "k", "markerfacecolor": "white"},
                        flierprops={"marker": "o", "markerfacecolor": "black", "markeredgecolor": "none", "markersize": 3},
                        boxprops={'edgecolor': 'k'}, width=.8, orient=orientation, **xy, hue=self.ahead_dim)

            _labels = [sampling_letter + str(i) for i in data_table.index.levels[1].values]
            handles, _ = ax.get_legend_handles_labels()
            ax.legend(handles, _labels)
        else:
            xy = (self.model_type_dim, 0) if orientation == "v" else (0, self.model_type_dim)
            col_or_row = {"col": self.ahead_dim} if orientation == "v" else {"row": self.ahead_dim}
            aspect = figsize[0] / figsize[1]
            height = figsize[1] * 0.8
            ax = sns.FacetGrid(data_table.stack(self.model_type_dim).reset_index(), **col_or_row, aspect=aspect, height=height)
            ax.map(sns.boxplot, *xy, whis=1.5, color="white", showmeans=True, order=data_table.mean().index.to_list(),
                   meanprops={"markersize": 6, "markeredgecolor": "k"},
                   flierprops={"marker": "o", "markerfacecolor": "black", "markeredgecolor": "none", "markersize": 3},
                   boxprops={'facecolor': 'none', 'edgecolor': 'k'}, width=width, orient=orientation)

            _labels = [sampling_letter + str(i) for i in data_table.index.levels[1].values]
            for axi, title in zip(ax.axes.flatten(), _labels):
                axi.set_title(title)
                plt.setp(axi.lines, color='k')

        error_label = self.error_measure if self.error_unit is None else f"{self.error_measure} (in {self.error_unit})"
        if agg_type == "panel":
            if orientation == "v":
                for axi in ax.axes.flatten():
                    axi.set_xlabel(None)
                    axi.set_xticklabels(axi.get_xticklabels(), rotation=45)
                ax.set_ylabels(error_label)
                loc = "upper left"
            else:
                for axi in ax.axes.flatten():
                    axi.set_ylabel(None)
                ax.set_xlabels(error_label)
                loc = "upper right"
            text = f"n={n_boots}"
            if self.block_length is not None:
                text = f"{self.block_length}, {text}"
            if len(season) > 0:
                text = f"{season}, {text}"
            text_box = AnchoredText(text, frameon=True, loc=loc)
            plt.setp(text_box.patch, edgecolor='k', facecolor='w')
            ax.axes.flatten()[0].add_artist(text_box)
        else:
            if orientation == "v":
                if apply_u_test:
                    ax = self.set_significance_bars(asteriks, ax, data_table, orientation)
                ylims = list(ax.get_ylim())
                ax.set_ylim([ylims[0], ylims[1]*1.025])
                ax.set_ylabel(error_label)
                ax.set_xticklabels(ax.get_xticklabels(), rotation=45)
                ax.set_xlabel(None)
            elif orientation == "h":
                if apply_u_test:
                    ax = self.set_significance_bars(asteriks, ax, data_table, orientation)
                ax.set_xlabel(error_label)
                xlims = list(ax.get_xlim())
                ax.set_xlim([xlims[0], xlims[1] * 1.015])
                ax.set_ylabel(None)
            else:
                raise ValueError(f"orientation must be `v' or `h' but is: {orientation}")
            text = f"n={n_boots}"
            if self.block_length is not None:
                text = f"{self.block_length}, {text}"
            if len(season) > 0:
                text = f"{season}, {text}"
            loc = "lower left"
            text_box = AnchoredText(text, frameon=True, loc=loc, pad=0.5, bbox_to_anchor=(0., 1.0),
                                    bbox_transform=ax.transAxes)
            plt.setp(text_box.patch, edgecolor='k', facecolor='w')
            ax.add_artist(text_box)
            plt.setp(ax.lines, color='k')
        plt.tight_layout()
        self._save()
        plt.close("all")

    def set_significance_bars(self, asteriks, ax, data_table, orientation):
        p1 = list(asteriks.index).index(self.model_name)
        q_prev = 0.
        factor = 0.025
        for i, ast in enumerate(asteriks):
            if not i == list(asteriks.index).index(self.model_name):
                p2 = i
                q = data_table[[self.model_name, data_table.columns[i]]].max().max()
                q = max(q, q_prev) * (1 + factor)
                if abs(q - q_prev) < q * factor:
                    q = q * (1 + factor)
                h = 0.01 * data_table.max().max()
                if orientation == "h":
                    ax.plot([q, q + h, q + h, q], [p1, p1, p2, p2], c="k")
                    ax.text(q + h, (p1 + p2) * 0.5, ast, ha="left", va="center", color="k", rotation=-90)
                elif orientation == "v":
                    ax.plot([p1, p1, p2, p2], [q, q + h, q + h, q], c="k")
                    ax.text((p1 + p2) * 0.5, q + h, ast, ha="center", va="bottom", color="k")
                q_prev = q
        return ax


@TimeTrackingWrapper
class PlotTimeEvolutionMetric(AbstractPlotClass):

    def __init__(self, data: xr.DataArray, ahead_dim="ahead", model_type_dim="type", plot_folder=".",
                 error_measure: str = "mse", error_unit: str = None, model_name: str = "NN",
                 model_indicator: str = "nn", time_dim="index"):
        super().__init__(plot_folder, "time_evolution_mse")
        self.title = error_measure + f" (in {error_unit})" if error_unit is not None else ""
        plot_name = self.plot_name
        vmin = int(data.quantile(0.05))
        vmax = int(data.quantile(0.95))
        data = self._prepare_data(data, time_dim, model_type_dim, model_indicator, model_name)

        # detailed plot for each model type
        for t in data[model_type_dim]:
            # note: could be expanded to create plot per ahead step
            plot_data = data.sel({model_type_dim: t}).mean(ahead_dim).to_pandas()
            years = plot_data.columns.strftime("%Y").to_list()
            months = plot_data.columns.strftime("%b").to_list()
            plot_data.columns = plot_data.columns.strftime("%b %Y")
            self.plot_name = f"{plot_name}_{t.values}"
            self._plot(plot_data, years, months, vmin, vmax, str(t.values))

        # aggregated version with all model types
        remaining_dim = set(data.dims).difference((model_type_dim, time_dim))
        _data = data.mean(remaining_dim, skipna=True).transpose(model_type_dim, time_dim)
        vmin = int(_data.quantile(0.05))
        vmax = int(_data.quantile(0.95))
        plot_data = _data.to_pandas()
        years = plot_data.columns.strftime("%Y").to_list()
        months = plot_data.columns.strftime("%b").to_list()
        plot_data.columns = plot_data.columns.strftime("%b %Y")
        self.plot_name = f"{plot_name}_summary"
        self._plot(plot_data, years, months, vmin, vmax, None)

        # line plot version
        y_dim = "error"
        plot_data = data.to_dataset(name=y_dim).to_dataframe().reset_index()
        self.plot_name = f"{plot_name}_line_plot"
        self._plot_summary_line(plot_data, x_dim=time_dim, y_dim=y_dim, hue_dim=model_type_dim)

    @staticmethod
    def _find_nan_edge(data, time_dim):
        coll = []
        for i in data:
            if bool(i) is False:
                break
            else:
                coll.append(i[time_dim].values)
        return coll

    def _prepare_data(self, data, time_dim, model_type_dim, model_indicator, model_name):
        # remove nans at begin and end
        nan_locs = data.isnull().all(helpers.remove_items(data.dims, time_dim))
        nans_at_end = self._find_nan_edge(reversed(nan_locs), time_dim)
        nans_at_begin = self._find_nan_edge(nan_locs, time_dim)
        data = data.drop(nans_at_begin + nans_at_end, time_dim)
        # rename nn model
        data[model_type_dim] = [v if v != model_indicator else model_name for v in data[model_type_dim].data.tolist()]
        return data

    @staticmethod
    def _set_ticks(ax, years, months):
        from matplotlib.ticker import IndexLocator
        ax.xaxis.set_major_locator(IndexLocator(1, 0.5))
        locs = ax.get_xticks(minor=False).tolist()[:len(months)]
        ax.set_xticks(locs, minor=True)
        ax.set_xticklabels([m[0] for m in months], minor=True, rotation=0)
        locs_major = []
        labels_major = []
        for l, major, minor in zip(locs, years, months):
            if minor == "Jan":
                locs_major.append(l + 0.001)
                labels_major.append(major)
        if len(locs_major) == 0:  # in case there is less than a year and no Jan included
            locs_major = locs[0] + 0.001
            labels_major = years[0]
        ax.set_xticks(locs_major)
        ax.set_xticklabels(labels_major, minor=False, rotation=0)
        ax.tick_params(axis="x", which="major", pad=15)

    @staticmethod
    def _aspect_cbar(val):
        return min(max(1.25 * val + 7.5, 5), 30)

    def _plot(self, data, years, months, vmin=None, vmax=None, subtitle=None):
        fig, ax = plt.subplots(figsize=(max(data.shape[1] / 6, 12), max(data.shape[0] / 3.5, 2)))
        data.sort_index(inplace=True)
        sns.heatmap(data, linewidths=1, cmap="coolwarm", ax=ax, vmin=vmin, vmax=vmax,
                    cbar_kws={"aspect": self._aspect_cbar(data.shape[0])})
        # or cmap="Spectral_r", cmap="RdYlBu_r", cmap="coolwarm",
        # square=True
        self._set_ticks(ax, years, months)
        ax.set(xlabel=None, ylabel=None, title=self.title if subtitle is None else f"{subtitle}\n{self.title}")
        plt.tight_layout()
        self._save()

    def _plot_summary_line(self, data, x_dim, y_dim, hue_dim):
        data[x_dim] = pd.to_datetime(data[x_dim].dt.strftime('%Y-%m')) #???
        n = len(data[hue_dim].unique())
        ax = sns.lineplot(data=data, x=x_dim, y=y_dim, hue=hue_dim, errorbar=("pi", 50),
                          palette=sns.color_palette()[:n], style=hue_dim, dashes=False, markers=["X"]*n)
        ax.set(xlabel=None, ylabel=self.title)
        ax.get_legend().set_title(None)
        ax.xaxis.set_major_locator(mdates.YearLocator())
        ax.xaxis.set_major_formatter(mdates.DateFormatter('%b\n%Y'))
        ax.xaxis.set_minor_locator(mdates.MonthLocator(bymonth=range(1, 13, 3)))
        ax.xaxis.set_minor_formatter(mdates.DateFormatter('%b'))
        ax.margins(x=0)
        plt.tight_layout()
        self._save()


@TimeTrackingWrapper
class PlotSeasonalMSEStack(AbstractPlotClass):

    def __init__(self, data, data_path: str, plot_folder: str = ".", boot_dim="boots", ahead_dim="ahead",
                 sampling: str = "daily", error_measure: str = "MSE", error_unit: str = "ppb$^2$", time_dim="index",
                 model_type_dim: str = "type", model_name: str = "NN", model_indicator: str = "nn",):
        """Set attributes and create plot."""
        super().__init__(plot_folder, "seasonal_mse_stack_plot")
        self._data_path = data_path
        self.season_dim = "season"
        self.time_dim = time_dim
        self.ahead_dim = ahead_dim
        self.error_unit = error_unit
        self.error_measure = error_measure
        self.dim_order = [self.season_dim, ahead_dim, model_type_dim]

        # mse from monthly blocks
        self.plot_name_orig = "seasonal_mse_stack_plot"
        self._data = self._prepare_data(data)
        for orientation in ["horizontal", "vertical"]:
            for split_ahead in [True, False]:
                self._plot(ahead_dim, split_ahead, sampling, orientation)
                self._save(bbox_inches="tight")

        # mes from resampling
        self.plot_name_orig = "seasonal_mse_from_uncertainty_stack_plot"
        self._data = self._prepare_data_from_uncertainty(boot_dim, data_path, model_type_dim, model_indicator,
                                                         model_name)
        for orientation in ["horizontal", "vertical"]:
            for split_ahead in [True, False]:
                self._plot(ahead_dim, split_ahead, sampling, orientation)
                self._save(bbox_inches="tight")

    def _prepare_data(self, data):
        season_mean = data.groupby(f"{self.time_dim}.{self.season_dim}").mean()
        total_mean = data.mean(self.time_dim)
        factor = season_mean / season_mean.sum(self.season_dim)
        season_share = (total_mean * factor).reindex({self.season_dim: ["DJF", "MAM", "JJA", "SON"]})
        season_share = season_share.mean(set(season_share.dims).difference(self.dim_order))
        return season_share.sortby(season_share.sum([self.season_dim, self.ahead_dim])).transpose(*self.dim_order)

    def _prepare_data_from_uncertainty(self, boot_dim, data_path, model_type_dim, model_indicator, model_name):
        season_dim = self.season_dim
        data = {}
        for season in ["total", "DJF", "MAM", "JJA", "SON"]:
            if season == "total":
                file_name = "uncertainty_estimate_raw_results.nc"
            else:
                file_name = f"uncertainty_estimate_raw_results_{season}.nc"
            with xr.open_dataarray(os.path.join(data_path, file_name)) as d:
                data[season] = d
        mean = {}
        for season in data.keys():
            mean[season] = data[season].mean(boot_dim)
        xr_data = xr.Dataset(mean).to_array(season_dim)
        xr_data[model_type_dim] = [v if v != model_indicator else model_name for v in xr_data[model_type_dim].values]
        xr_season = xr_data.sel({season_dim: ["DJF", "MAM", "JJA", "SON"]})
        factor = xr_season / xr_season.sum(season_dim)
        season_share = xr_data.sel({season_dim: "total"}) * factor
        return season_share.sortby(season_share.sum([self.season_dim, self.ahead_dim])).transpose(*self.dim_order)

    @staticmethod
    def _set_bar_label(ax):
        opts = {}
        sum = {}
        for c in ax.containers:
            labels = [v for v in c.datavalues]
            opts[c] = labels
            sum = {i: sum.get(i, 0) + l for (i, l) in enumerate(labels)}
        for c, labels in opts.items():
            _l = [f"{round(100 * labels[i] / sum[i])}%" for i in range(len(labels))]
            ax.bar_label(c, labels=_l, label_type='center')

    def _plot(self, dim, split_ahead=True, sampling="daily", orientation="vertical"):
        _, sampling_letter = self._get_sampling(sampling, 1)
        sampling_letter = {"d": "D", "h": "H"}.get(sampling_letter, sampling_letter)
        if split_ahead is False:
            self.plot_name = self.plot_name_orig + "_total_" + orientation
            data = self._data.mean(dim)
            if orientation == "vertical":
                fig, ax = plt.subplots(1, 1)
                data.to_pandas().T.plot.bar(ax=ax, stacked=True, cmap="Dark2", legend=False)
                ax.xaxis.label.set_visible(False)
                ax.set_ylabel(f"{self.error_measure} (in {self.error_unit})")
                self._set_bar_label(ax)
            else:
                m = data.to_pandas().T.shape[0]
                fig, ax = plt.subplots(1, 1, figsize=(6, m))
                data.to_pandas().T.plot.barh(ax=ax, stacked=True, cmap="Dark2", legend=False)
                ax.yaxis.label.set_visible(False)
                ax.set_xlabel(f"{self.error_measure} (in {self.error_unit})")
                self._set_bar_label(ax)
            fig.legend(*ax.get_legend_handles_labels(), loc="upper center", ncol=4)
            fig.tight_layout(rect=[0, 0, 1, 0.9])
        else:
            self.plot_name = self.plot_name_orig + "_" + orientation
            data = self._data
            n = len(data.coords[dim])
            m = data.max(self.season_dim).shape
            if orientation == "vertical":
                fig, ax = plt.subplots(1, n, sharey=True, figsize=(np.prod(m) / 0.8, 5))
                for i, sel in enumerate(data.coords[dim].values):
                    data.sel({dim: sel}).to_pandas().T.plot.bar(ax=ax[i], stacked=True, cmap="Dark2", legend=False)
                    label = sampling_letter + str(sel)
                    ax[i].set_title(label)
                    ax[i].xaxis.label.set_visible(False)
                    self._set_bar_label(ax[i])
                ax[0].set_ylabel(f"{self.error_measure} (in {self.error_unit})")
                fig.legend(*ax[0].get_legend_handles_labels(), loc="upper center", ncol=4)
                fig.tight_layout(rect=[0, 0, 1, 0.9])
            else:
                fig, ax = plt.subplots(n, 1, sharex=True, figsize=(6, np.prod(m) * 0.6))
                for i, sel in enumerate(data.coords[dim].values):
                    data.sel({dim: sel}).to_pandas().T.plot.barh(ax=ax[i], stacked=True, cmap="Dark2", legend=False)
                    label = sampling_letter + str(sel)
                    ax[i].set_title(label)
                    ax[i].yaxis.label.set_visible(False)
                    self._set_bar_label(ax[i])
                ax[-1].set_xlabel(f"{self.error_measure} (in {self.error_unit})")
                fig.legend(*ax[0].get_legend_handles_labels(), loc="upper center", ncol=4)
                fig.tight_layout(rect=[0, 0, 1, 0.95])


@TimeTrackingWrapper
class PlotErrorsOnMap(AbstractPlotClass):
    from mlair.plotting.data_insight_plotting import PlotStationMap

    def __init__(self, data_gen, errors, error_metric, plot_folder: str = ".", iter_dim: str = "station",
                 model_type_dim: str = "type", ahead_dim: str = "ahead", sampling: str = "daily"):

        super().__init__(plot_folder, f"map_plot_{error_metric}")
        plot_path = os.path.join(self.plot_folder, f"{self.plot_name}.pdf")
        pdf_pages = matplotlib.backends.backend_pdf.PdfPages(plot_path)
        error_metric_units = helpers.statistics.get_error_metrics_units("ppb")[error_metric]
        error_metric_name = helpers.statistics.get_error_metrics_long_name()[error_metric]
        self.sampling = self._get_sampling(sampling, 1)[1]

        coords = self._extract_coords(data_gen)
        for split_ahead in [False, True]:
            error_data = {}
            for model_type in errors.coords[model_type_dim].values:
                error_data[model_type] = self._prepare_data(errors, model_type_dim, model_type, ahead_dim, error_metric,
                                                            split_ahead=split_ahead)
            limits = self._calculate_limits(error_data)
            for model_type, error in error_data.items():
                if split_ahead is True:
                    for ahead in error.index.unique(1).to_list():
                        error_ahead = error.query(f"{ahead_dim} == {ahead}").droplevel(1)
                        plot_data = pd.concat([coords, error_ahead], axis=1)
                        self.plot(plot_data, error_metric, error_metric_name, error_metric_units, model_type, limits,
                                  ahead=ahead)
                        pdf_pages.savefig()
                else:
                    plot_data = pd.concat([coords, error], axis=1)
                    self.plot(plot_data, error_metric, error_metric_name, error_metric_units, model_type, limits)
                    pdf_pages.savefig()
        pdf_pages.close()
        plt.close('all')

    @staticmethod
    def _calculate_limits(data):
        vmin, vmax = np.inf, -np.inf
        for v in data.values():
            vmin = min(vmin, v.min().values)
            vmax = max(vmax, v.max().values)
        return relative_round(float(vmin), 2, floor=True), relative_round(float(vmax), 2, ceil=True)

    @staticmethod
    def _set_bounds(limits, ncolors, error_metric):
        bound_lims = {"ioa": [0, 1], "mnmb": [-2, 2]}.get(error_metric, limits)
        vmin = relative_round(bound_lims[0], 2, floor=True)
        vmax = relative_round(bound_lims[1], 2, ceil=True)
        interval = relative_round((vmax - vmin) / ncolors, 1, ceil=True)
        bounds = np.sort(np.arange(vmax, vmin, -interval))
        return bounds

    @staticmethod
    def _get_colorpalette(error_metric):
        # cmap = matplotlib.cm.coolwarm
        # cmap = sns.color_palette("magma_r", as_cmap=True)
        # cmap="Spectral_r", cmap="RdYlBu_r", cmap="coolwarm",
        # cmap = sns.cubehelix_palette(8, start=2, rot=0, dark=0, light=.95, as_cmap=True)
        if error_metric == "mnmb":
            cmap = sns.mpl_palette("coolwarm", as_cmap=True)
        elif error_metric == "ioa":
            cmap = sns.mpl_palette("coolwarm_r", as_cmap=True)
        else:
            cmap = sns.color_palette("magma_r", as_cmap=True)
        return cmap

    def plot(self, plot_data, error_metric, error_long_name, error_units, model_type, limits, ahead=None):
        import cartopy.crs as ccrs
        from cartopy.mpl.gridliner import LONGITUDE_FORMATTER, LATITUDE_FORMATTER
        fig = plt.figure(figsize=(10, 5))
        ax = fig.add_subplot(1, 1, 1, projection=ccrs.PlateCarree())
        _gl = ax.gridlines(xlocs=range(-180, 180, 5), ylocs=range(-90, 90, 2), draw_labels=True)
        _gl.xformatter = LONGITUDE_FORMATTER
        _gl.yformatter = LATITUDE_FORMATTER
        self._draw_background(ax)
        cmap = self._get_colorpalette(error_metric)
        ncolors = 20
        bounds = self._set_bounds(limits, ncolors, error_metric)
        norm = colors.BoundaryNorm(bounds, cmap.N, extend='both')
        cb = ax.scatter(plot_data["lon"], plot_data["lat"], c=plot_data[error_metric], marker='o', s=50,
                        transform=ccrs.PlateCarree(), zorder=2, cmap=cmap, norm=norm)
        cbar_label = f"{error_long_name} (in {error_units})" if error_units is not None else error_long_name
        plt.colorbar(cb, label=cbar_label)
        self._adjust_extent(ax)
        sampling_letter = {"d": "D", "h": "H"}.get(self.sampling, self.sampling)
        title = model_type if ahead is None else f"{model_type} ({sampling_letter}{ahead})"
        plt.title(title)
        plt.tight_layout()

    @staticmethod
    def _adjust_extent(ax):
        import cartopy.crs as ccrs

        def diff(arr):
            return arr[1] - arr[0], arr[3] - arr[2]

        def find_ratio(delta, reference=5):
            return min(max(abs(reference / delta[0]), abs(reference / delta[1])), 5)

        extent = ax.get_extent(crs=ccrs.PlateCarree())
        ratio = find_ratio(diff(extent))
        new_extent = extent + np.array([-1, 1, -1, 1]) * ratio
        ax.set_extent(new_extent, crs=ccrs.PlateCarree())

    @staticmethod
    def _extract_coords(gen):
        coll = []
        for station in gen:
            coords = station.get_coordinates()
            coll.append((str(station), coords["lon"], coords["lat"]))
        return pd.DataFrame(coll, columns=["station", "lon", "lat"]).set_index("station")

    @staticmethod
    def _prepare_data(errors, model_type_dim, model_type, ahead_dim, error_metric, split_ahead=False):
        e = errors.sel({model_type_dim: model_type}, drop=True)
        if split_ahead is False:
            e = e.mean(ahead_dim)
        return e.to_dataframe(error_metric)

    @staticmethod
    def _draw_background(ax):
        """Draw coastline, lakes, ocean, rivers and country borders as background on the map."""

        import cartopy.feature as cfeature

        ax.add_feature(cfeature.LAND.with_scale("50m"))
        ax.natural_earth_shp(resolution='50m')
        ax.add_feature(cfeature.COASTLINE.with_scale("50m"), edgecolor='black')
        ax.add_feature(cfeature.LAKES.with_scale("50m"))
        ax.add_feature(cfeature.OCEAN.with_scale("50m"))
        ax.add_feature(cfeature.RIVERS.with_scale("50m"))
        ax.add_feature(cfeature.BORDERS.with_scale("50m"), facecolor='none', edgecolor='black')






    def _plot_individual(self):
        import cartopy.feature as cfeature
        import cartopy.crs as ccrs
        from cartopy.mpl.gridliner import LONGITUDE_FORMATTER, LATITUDE_FORMATTER
        from mpl_toolkits.axes_grid1 import make_axes_locatable

        for competitor in self.reference_models:
            file_name = os.path.join(self.skill_score_report_path,
                                     f"error_report_skill_score_{self.model_name}_-_{competitor}.csv"
                                     )

            plot_path = os.path.join(os.path.abspath(self.plot_folder),
                                     f"{self.plot_name}_{self.model_name}_-_{competitor}.pdf")
            pdf_pages = matplotlib.backends.backend_pdf.PdfPages(plot_path)

            for i, lead_name in enumerate(df.columns[:-2]):  # last two are lat lon
                fig = plt.figure()
                self._ax.scatter(df.lon.values, df.lat.values, c=df[lead_name],
                           transform=ccrs.PlateCarree(),
                           norm=norm, cmap=cmap)
                self._gl = self._ax.gridlines(xlocs=range(0, 21, 5), ylocs=range(44, 59, 2), draw_labels=True)
                self._gl.xformatter = LONGITUDE_FORMATTER
                self._gl.yformatter = LATITUDE_FORMATTER
                label = f"Skill Score: {lead_name.replace('-', 'vs.').replace('(t+', ' (').replace(')', 'd)')}"
                self._cbar = fig.colorbar(mpl.cm.ScalarMappable(norm=norm, cmap=cmap),
                                          orientation='horizontal', ticks=ticks,
                                          label=label,
                                          # cax=cax
                                          )

            # close all open figures / plots
            pdf_pages.savefig()
            pdf_pages.close()
            plt.close('all')

    def _plot(self, ncol: int = 2):
        import cartopy.feature as cfeature
        import cartopy.crs as ccrs
        from cartopy.mpl.gridliner import LONGITUDE_FORMATTER, LATITUDE_FORMATTER
        import string
        base_plot_name = self.plot_name
        for competitor in self.reference_models:
            file_name = os.path.join(self.skill_score_report_path,
                                     f"error_report_skill_score_{self.model_name}_-_{competitor}.csv"
                                     )

            self.plot_name = f"{base_plot_name}_{self.model_name}_-_{competitor}"
            df = self.open_data(file_name)

            nrow = int(np.ceil(len(df.columns[:-2])/ncol))
            bounds = np.linspace(-1, 1, 100)
            cmap = mpl.cm.coolwarm
            norm = colors.BoundaryNorm(bounds, cmap.N, extend='both')
            ticks = np.arange(norm.vmin, norm.vmax + .2, .2)
            fig, self._axes = plt.subplots(nrows=nrow, ncols=ncol, subplot_kw={'projection': ccrs.PlateCarree()})
            for i, ax in enumerate(self._axes.reshape(-1)):  # last two are lat lon

                sub_name = f"({string.ascii_lowercase[i]})"
                lead_name = df.columns[i]
                ax.add_feature(cfeature.LAND.with_scale("50m"))
                ax.add_feature(cfeature.COASTLINE.with_scale("50m"), edgecolor='black')
                ax.add_feature(cfeature.OCEAN.with_scale("50m"))
                ax.add_feature(cfeature.RIVERS.with_scale("50m"))
                ax.add_feature(cfeature.BORDERS.with_scale("50m"), facecolor='none', edgecolor='black')
                ax.scatter(df.lon.values, df.lat.values, c=df[lead_name],
                           marker='.',
                           transform=ccrs.PlateCarree(),
                           norm=norm, cmap=cmap)
                gl = ax.gridlines(xlocs=range(0, 21, 5), ylocs=range(44, 59, 2), draw_labels=True)
                gl.xformatter = LONGITUDE_FORMATTER
                gl.yformatter = LATITUDE_FORMATTER
                gl.top_labels = []
                gl.right_labels = []
                ax.text(0.01, 1.09, f'{sub_name} {lead_name.split("+")[1][:-1]}d',
                        verticalalignment='top', horizontalalignment='left',
                        transform=ax.transAxes,
                        color='black',
                        )
            label = f"Skill Score: {lead_name.replace('-', 'vs.').split('(')[0]}"

            fig.subplots_adjust(bottom=0.18)
            cax = fig.add_axes([0.15, 0.1, 0.7, 0.02])
            self._cbar = fig.colorbar(mpl.cm.ScalarMappable(norm=norm, cmap=cmap),
                                      orientation='horizontal',
                                      ticks=ticks,
                                      label=label,
                                      cax=cax
                                      )

            fig.subplots_adjust(wspace=.001, hspace=.2)
            self._save(bbox_inches="tight")
            plt.close('all')


    @staticmethod
    def get_coords_from_index(name_string: str) -> List[float]:
        """

        :param name_string:
        :type name_string:
        :return: List of coords [lat, lon]
        :rtype: List
        """
        res = [float(frac.replace("_", ".")) for frac in name_string.split(sep="__")[1:]]
        return res
