"""Pre-processing module."""

__author__ = "Lukas Leufen, Felix Kleinert"
__date__ = '2019-11-25'

import logging
import os
import traceback
from typing import Tuple
import multiprocessing
import requests
import psutil
import random
import dill

import pandas as pd

from mlair.data_handler import DataCollection, AbstractDataHandler
from mlair.helpers import TimeTracking, to_list, tables, remove_items
from mlair.configuration import path_config
from mlair.helpers.data_sources.data_loader import EmptyQueryResult
from mlair.helpers.testing import check_nested_equality
from mlair.run_modules.run_environment import RunEnvironment


class PreProcessing(RunEnvironment):
    """
    Pre-process your data by using this class.

    Schedule of pre-processing:
        #. load and check valid stations (either download or load from disk)
        #. split subsets (train, val, test, train & val)
        #. create small report on data metrics

    Required objects [scope] from data store:
        * all elements from `DEFAULT_ARGS_LIST` in scope preprocessing for general data loading
        * all elements from `DEFAULT_ARGS_LIST` in scopes [train, val, test, train_val] for custom subset settings
        * `fraction_of_training` [.]
        * `experiment_path` [.]
        * `use_all_stations_on_all_data_sets` [.]

    Optional objects
        * all elements from `DEFAULT_KWARGS_LIST` in scope preprocessing for general data loading
        * all elements from `DEFAULT_KWARGS_LIST` in scopes [train, val, test, train_val] for custom subset settings

    Sets
        * `stations` in [., train, val, test, train_val]
        * `generator` in [train, val, test, train_val]
        * `transformation` [.]

    Creates
        * all input and output data in `data_path`
        * latex reports in `experiment_path/latex_report`

    """

    def __init__(self):
        """Set up and run pre-processing."""
        super().__init__()
        self._run()

    def _run(self):
        snapshot_load_path = self.data_store.get_default("snapshot_load_path", default=None)
        if snapshot_load_path is None:
            stations = self.data_store.get("stations")
            data_handler = self.data_store.get("data_handler")
            self._load_apriori()
            _, valid_stations = self.validate_station(data_handler, stations,
                                                      "preprocessing")  # , store_processed_data=False)
            if len(valid_stations) == 0:
                raise ValueError("Couldn't find any valid data according to given parameters. Abort experiment run.")
            self.data_store.set("stations", valid_stations)
            self.split_train_val_test()
        else:
            self.load_snapshot(snapshot_load_path)
        self.report_pre_processing()
        self.prepare_competitors()
        if self.data_store.get_default("create_snapshot", False) is True:
            self.create_snapshot()

    def report_pre_processing(self):
        """Log some metrics on data and create latex report."""
        logging.debug(20 * '##')
        n_train = len(self.data_store.get('data_collection', 'train'))
        n_val = len(self.data_store.get('data_collection', 'val'))
        n_test = len(self.data_store.get('data_collection', 'test'))
        n_total = n_train + n_val + n_test
        logging.debug(f"Number of all stations: {n_total}")
        logging.debug(f"Number of training stations: {n_train}")
        logging.debug(f"Number of val stations: {n_val}")
        logging.debug(f"Number of test stations: {n_test}")
        self.create_latex_report()

    def create_latex_report(self):
        """
        Create tables with information on the station meta data and a summary on subset sample sizes.

        * station_sample_size.md: see table below as markdown
        * station_sample_size.tex: same as table below as latex table
        * station_sample_size_short.tex: reduced size table without any meta data besides station ID, as latex table

        All tables are stored inside experiment_path inside the folder latex_report. The table format (e.g. which meta
        data is highlighted) is currently hardcoded to have a stable table style. If further styles are needed, it is
        better to add an additional style than modifying the existing table styles.

        +------------+-------------------------------------------+---------------+---------------+---------------+---------+-------+--------+
        | stat. ID   | station_name                              |   station_lon |   station_lat |   station_alt |   train |   val |   test |
        +============+===========================================+===============+===============+===============+=========+=======+========+
        | DEBW013    | Stuttgart Bad Cannstatt                   |        9.2297 |       48.8088 |           235 |    1434 |   712 |   1080 |
        +------------+-------------------------------------------+---------------+---------------+---------------+---------+-------+--------+
        | DEBW076    | Baden-Baden                               |        8.2202 |       48.7731 |           148 |    3037 |   722 |    710 |
        +------------+-------------------------------------------+---------------+---------------+---------------+---------+-------+--------+
        | DEBW087    | Schwäbische_Alb                           |        9.2076 |       48.3458 |           798 |    3044 |   714 |   1087 |
        +------------+-------------------------------------------+---------------+---------------+---------------+---------+-------+--------+
        | DEBW107    | Tübingen                                  |        9.0512 |       48.5077 |           325 |    1803 |   715 |   1087 |
        +------------+-------------------------------------------+---------------+---------------+---------------+---------+-------+--------+
        | DEBY081    | Garmisch-Partenkirchen/Kreuzeckbahnstraße |       11.0631 |       47.4764 |           735 |    2935 |   525 |    714 |
        +------------+-------------------------------------------+---------------+---------------+---------------+---------+-------+--------+
        | # Stations | nan                                       |      nan      |      nan      |           nan |       6 |     6 |      6 |
        +------------+-------------------------------------------+---------------+---------------+---------------+---------+-------+--------+
        | # Samples  | nan                                       |      nan      |      nan      |           nan |   12253 |  3388 |   4678 |
        +------------+-------------------------------------------+---------------+---------------+---------------+---------+-------+--------+

        """
        meta_cols = ["name", "lat", "lon", "alt", "country", "state", "type", "type_of_area", "toar1_category"]
        meta_round = ["lat", "lon", "alt"]
        precision = 4
        path = os.path.join(self.data_store.get("experiment_path"), "latex_report")
        path_config.check_path_and_create(path)
        names_of_set = ["train", "val", "test"]
        df = self.create_info_df(meta_cols, meta_round, names_of_set, precision)
        column_format = tables.create_column_format_for_tex(df)
        tables.save_to_tex(path=path, filename="station_sample_size.tex", column_format=column_format, df=df)
        tables.save_to_md(path=path, filename="station_sample_size.md", df=df)
        df_nometa = df.drop(meta_cols, axis=1)
        column_format = tables.create_column_format_for_tex(df)
        tables.save_to_tex(path=path, filename="station_sample_size_short.tex", column_format=column_format,
                           df=df_nometa)
        tables.save_to_md(path=path, filename="station_sample_size_short.md", df=df_nometa)
        df_descr = self.create_describe_df(df_nometa)
        column_format = tables.create_column_format_for_tex(df_descr)
        tables.save_to_tex(path=path, filename="station_describe_short.tex", column_format=column_format, df=df_descr)
        tables.save_to_md(path=path, filename="station_describe_short.md", df=df_descr)

    @staticmethod
    def create_describe_df(df, percentiles=None, ignore_last_lines: int = 2):
        if percentiles is None:
            percentiles = [.05, .1, .25, .5, .75, .9, .95]
        df_descr = df.iloc[:-ignore_last_lines].astype('float32').describe(
            percentiles=percentiles).astype("int32", errors="ignore")
        df_descr = pd.concat([df.loc[['# Samples']], df_descr]).T
        df_descr.rename(columns={"# Samples": "no. samples", "count": "no. stations"}, inplace=True)
        df_descr_colnames = list(df_descr.columns)
        df_descr_colnames = [df_descr_colnames[1]] + [df_descr_colnames[0]] + df_descr_colnames[2:]
        df_descr = df_descr[df_descr_colnames]
        return df_descr

    def create_info_df(self, meta_cols, meta_round, names_of_set, precision):
        use_multiprocessing = self.data_store.get("use_multiprocessing")
        max_process = self.data_store.get("max_number_multiprocessing")
        df = pd.DataFrame(columns=meta_cols + names_of_set)
        for set_name in names_of_set:
            data = self.data_store.get("data_collection", set_name)
            n_process = min([psutil.cpu_count(logical=False), len(data), max_process])  # use only physical cpus
            if n_process > 1 and use_multiprocessing is True:  # parallel solution
                logging.info(f"use parallel create_info_df ({set_name})")
                pool = multiprocessing.Pool(n_process)
                logging.info(f"running {getattr(pool, '_processes')} processes in parallel")
                output = [pool.apply_async(f_proc_create_info_df, args=(station, meta_cols)) for station in data]
                for i, p in enumerate(output):
                    res = p.get()
                    station_name, shape, meta = res["station_name"], res["Y_shape"], res["meta"]
                    df.loc[station_name, set_name] = shape
                    if df.loc[station_name, meta_cols].isnull().any():
                        df.loc[station_name, meta_cols] = meta
                    logging.info(f"...finished: {station_name} ({int((i + 1.) / len(output) * 100)}%)")
                pool.close()
                pool.join()
            else:  # serial solution
                logging.info(f"use serial create_info_df ({set_name})")
                for station in data:
                    res = f_proc_create_info_df(station, meta_cols)
                    station_name, shape, meta = res["station_name"], res["Y_shape"], res["meta"]
                    df.loc[station_name, set_name] = shape
                    if df.loc[station_name, meta_cols].isnull().any():
                        df.loc[station_name, meta_cols] = meta
            df.loc["# Samples", set_name] = df.loc[:, set_name].sum()
            assert len(data) == df.loc[:, set_name].count() - 1
            df.loc["# Stations", set_name] = len(data)
        df[meta_round] = df[meta_round].astype(float).round(precision)
        df.sort_index(inplace=True)
        df = df.reindex(df.index.drop(["# Stations", "# Samples"]).to_list() + ["# Stations", "# Samples"], )
        df.index.name = 'stat. ID'
        return df

    def split_train_val_test(self) -> None:
        """
        Split data into subsets.

        Currently: train, val, test and train_val (actually this is only the merge of train and val, but as an separate
        data_collection). IMPORTANT: Do not change to order of the execution of create_set_split. The train subset needs
        always to be executed at first, to set a proper transformation.
        """
        fraction_of_training = self.data_store.get("fraction_of_training")
        stations = self.data_store.get("stations")
        train_index, val_index, test_index, train_val_index = self.split_set_indices(len(stations),
                                                                                     fraction_of_training)
        subset_names = ["train", "val", "test", "train_val"]
        if subset_names[0] != "train":  # pragma: no cover
            raise AssertionError(f"Make sure, that the train subset is always at first execution position! Given subset"
                                 f"order was: {subset_names}.")
        for (ind, scope) in zip([train_index, val_index, test_index, train_val_index], subset_names):
            self.create_set_split(ind, scope)

    @staticmethod
    def split_set_indices(total_length: int, fraction: float) -> Tuple[slice, slice, slice, slice]:
        """
        Create the training, validation and test subset slice indices for given total_length.

        The test data consists on (1-fraction) of total_length (fraction*len:end). Train and validation data therefore
        are made from fraction of total_length (0:fraction*len). Train and validation data is split by the factor 0.8
        for train and 0.2 for validation. In addition, split_set_indices returns also the combination of training and
        validation subset.

        :param total_length: list with all objects to split
        :param fraction: ratio between test and union of train/val data

        :return: slices for each subset in the order: train, val, test, train_val
        """
        pos_test_split = int(total_length * fraction)
        train_index = slice(0, int(pos_test_split * 0.8))
        val_index = slice(int(pos_test_split * 0.8), pos_test_split)
        test_index = slice(pos_test_split, total_length)
        train_val_index = slice(0, pos_test_split)
        return train_index, val_index, test_index, train_val_index

    def create_set_split(self, index_list: slice, set_name: str) -> None:
        # get set stations
        stations = self.data_store.get("stations", scope=set_name)
        if self.data_store.get("use_all_stations_on_all_data_sets"):
            set_stations = stations
        else:
            set_stations = stations[index_list]
        logging.debug(f"{set_name.capitalize()} stations (len={len(set_stations)}): {set_stations}")
        # create set data_collection and store
        data_handler = self.data_store.get("data_handler")
        collection, valid_stations = self.validate_station(data_handler, set_stations, set_name)
        self.data_store.set("stations", valid_stations, scope=set_name)
        self.data_store.set("data_collection", collection, scope=set_name)

    def validate_station(self, data_handler: AbstractDataHandler, set_stations, set_name=None,
                         store_processed_data=True):
        """
        Check if all given stations in `all_stations` are valid.

        Valid means, that there is data available for the given time range (is included in `kwargs`). The shape and the
        loading time are logged in debug mode.

        :return: Corrected list containing only valid station IDs.
        """
        t_outer = TimeTracking()
        logging.info(f"check valid stations started{' (%s)' % (set_name if set_name is not None else 'all')}")
        # calculate transformation using train data
        if set_name == "train":
            logging.info("setup transformation using train data exclusively")
            self.transformation(data_handler, set_stations)
        # start station check
        collection = DataCollection(name=set_name)
        valid_stations = []
        kwargs = self.data_store.create_args_dict(data_handler.requirements(skip_args="station"), scope=set_name)
        use_multiprocessing = self.data_store.get("use_multiprocessing")
        tmp_path = self.data_store.get("tmp_path")

        max_process = self.data_store.get("max_number_multiprocessing")
        n_process = min([psutil.cpu_count(logical=False), len(set_stations), max_process])  # use only physical cpus
        if n_process > 1 and use_multiprocessing is True:  # parallel solution
            logging.info("use parallel validate station approach")
            pool = multiprocessing.Pool(n_process)
            logging.info(f"running {getattr(pool, '_processes')} processes in parallel")
            kwargs.update({"tmp_path": tmp_path, "return_strategy": "reference"})
            output = [
                pool.apply_async(f_proc, args=(data_handler, station, set_name, store_processed_data), kwds=kwargs)
                for station in set_stations]
            for i, p in enumerate(output):
                _res_file, s = p.get()
                logging.info(f"...finished: {s} ({int((i + 1.) / len(output) * 100)}%)")
                with open(_res_file, "rb") as f:
                    dh = dill.load(f)
                os.remove(_res_file)
                if dh is not None:
                    collection.add(dh)
                    valid_stations.append(s)
            pool.close()
            pool.join()
        else:  # serial solution
            logging.info("use serial validate station approach")
            kwargs.update({"return_strategy": "result"})
            for i, station in enumerate(set_stations):
                dh, s = f_proc(data_handler, station, set_name, store_processed_data, **kwargs)
                if dh is not None:
                    collection.add(dh)
                    valid_stations.append(s)
                logging.info(f"...finished: {s} ({int((i + 1.) / len(set_stations) * 100)}%)")

        logging.info(f"run for {t_outer} to check {len(set_stations)} station(s). Found {len(collection)}/"
                     f"{len(set_stations)} valid stations ({set_name}).")
        if set_name == "train":
            self.store_data_handler_attributes(data_handler, collection)
        return collection, valid_stations

    def store_data_handler_attributes(self, data_handler, collection):
        store_attributes = data_handler.store_attributes()
        if len(store_attributes) > 0:
            logging.info(f"store following parameters ({len(store_attributes)}) requested by the data handler: "
                         f"{','.join(store_attributes)}")
            attrs = {}
            for dh in collection:
                station = str(dh)
                for k, v in dh.get_store_attributes().items():
                    attrs[k] = dict(attrs.get(k, {}), **{station: v})
            for k, v in attrs.items():
                self.data_store.set(k, v)
        self._store_apriori()

    def _store_apriori(self):
        apriori = self.data_store.get_default("apriori", default=None)
        if apriori:
            experiment_path = self.data_store.get("experiment_path")
            path = os.path.join(experiment_path, "data", "apriori")
            store_file = os.path.join(path, "apriori.pickle")
            if not os.path.exists(path):
                path_config.check_path_and_create(path)
            with open(store_file, "wb") as f:
                dill.dump(apriori, f, protocol=4)
            logging.debug(f"Store apriori options locally for later use at: {store_file}")

    def _load_apriori(self):
        if self.data_store.get_default("apriori", default=None) is None:
            apriori_file = self.data_store.get_default("apriori_file", None)
            if apriori_file is not None:
                if os.path.exists(apriori_file):
                    logging.info(f"use apriori data from given file: {apriori_file}")
                    with open(apriori_file, "rb") as pickle_file:
                        self.data_store.set("apriori", dill.load(pickle_file))
                else:
                    logging.info(f"cannot load apriori file: {apriori_file}. Use fresh calculation from data.")

    def transformation(self, data_handler: AbstractDataHandler, stations):
        calculate_fresh_transformation = self.data_store.get_default("calculate_fresh_transformation", True)
        if hasattr(data_handler, "transformation"):
            transformation_opts = None if calculate_fresh_transformation is True else self._load_transformation()
            if transformation_opts is None:
                logging.info(f"start to calculate transformation parameters.")
                kwargs = self.data_store.create_args_dict(data_handler.requirements(skip_args="station"), scope="train")
                tmp_path = self.data_store.get_default("tmp_path", default=None)
                transformation_opts = data_handler.transformation(stations, tmp_path=tmp_path, **kwargs)
            else:
                logging.info("In case no valid train data could be found due to problems with transformation, please "
                             "check your provided transformation file for compability with your data.")
            self.data_store.set("transformation", transformation_opts)
            if transformation_opts is not None:
                self._store_transformation(transformation_opts)

    def _load_transformation(self):
        """Try to load transformation options from file if transformation_file is provided."""
        transformation_file = self.data_store.get_default("transformation_file", None)
        if transformation_file is not None:
            if os.path.exists(transformation_file):
                logging.info(f"use transformation from given transformation file: {transformation_file}")
                with open(transformation_file, "rb") as pickle_file:
                    return dill.load(pickle_file)
            else:
                logging.info(f"cannot load transformation file: {transformation_file}. Use fresh calculation of "
                             f"transformation from train data.")

    def _store_transformation(self, transformation_opts):
        """Store transformation options locally inside experiment_path if not exists already."""
        experiment_path = self.data_store.get("experiment_path")
        transformation_path = os.path.join(experiment_path, "data", "transformation")
        transformation_file = os.path.join(transformation_path, "transformation.pickle")
        calculate_fresh_transformation = self.data_store.get_default("calculate_fresh_transformation", True)
        if not os.path.exists(transformation_file) or calculate_fresh_transformation:
            path_config.check_path_and_create(transformation_path)
            with open(transformation_file, "wb") as f:
                dill.dump(transformation_opts, f, protocol=4)
            logging.info(f"Store transformation options locally for later use at: {transformation_file}")

    def prepare_competitors(self):
        """
        Prepare competitor models already in the preprocessing stage. This is performed here, because some models might
        need to have internet access, which is depending on the operating system not possible during postprocessing.
        This method checks currently only, if the Intelli03-ts-v1 model is requested as competitor and downloads the
        data if required.
        """
        logging.info("Searching for competitors to be prepared for use.")
        competitors = to_list(self.data_store.get_default("competitors", default=[]))
        if len(competitors) > 0:
            for competitor_name in competitors:
                if competitor_name.lower() == "IntelliO3-ts-v1".lower():
                    logging.info("Prepare IntelliO3-ts-v1 model")
                    from mlair.reference_models.reference_model_intellio3_v1 import IntelliO3_ts_v1
                    path = os.path.join(self.data_store.get("competitor_path"), competitor_name)
                    IntelliO3_ts_v1("IntelliO3-ts-v1", ref_store_path=path).make_reference_available_locally(remove_tmp_dir=False)
                elif competitor_name.lower() == "CAMS".lower():
                    logging.info("Prepare CAMS forecasts")
                    from mlair.reference_models.reference_model_cams import CAMSforecast
                    interp_method = self.data_store.get_default("cams_interp_method", default=None)
                    data_path = self.data_store.get_default("cams_data_path", default=None)
                    path = os.path.join(self.data_store.get("competitor_path"), competitor_name)
                    stations = {}
                    for subset in ["train", "val", "test"]:
                        data_collection = self.data_store.get("data_collection", subset)
                        stations.update({str(s): s.get_coordinates() for s in data_collection if s not in stations})
                    if interp_method is None:
                        CAMSforecast("CAMS", ref_store_path=path, data_path=data_path, interp_method=None
                                     ).make_reference_available_locally(stations)
                    else:
                        competitors = remove_items(competitors, "CAMS")
                        for method in to_list(interp_method):
                            CAMSforecast(f"CAMS{method}", ref_store_path=path + method, data_path=data_path,
                                         interp_method=method).make_reference_available_locally(stations)
                            competitors.append(f"CAMS{method}")
                        self.data_store.set("competitors", competitors)
                else:
                    logging.info(f"No preparation required for competitor {competitor_name} as no specific instruction "
                                 f"is provided.")
        else:
            logging.info("No preparation required because no competitor was provided to the workflow.")

    def create_snapshot(self):
        logging.info("create snapshot for preprocessing")
        from mlair.configuration.snapshot_names import animals
        for i_try in range(10):
            snapshot_name = random.choice(animals).lower()
            snapshot_path = os.path.abspath(self.data_store.get("snapshot_path"))
            path_config.check_path_and_create(snapshot_path, remove_existing=False)
            _snapshot_file = os.path.join(snapshot_path, f"snapshot_preprocessing_{snapshot_name}.pickle")
            if not os.path.exists(_snapshot_file):
                logging.info(f"store snapshot at: {_snapshot_file}")
                with open(_snapshot_file, "wb") as f:
                    dill.dump(self.data_store, f, protocol=4)
                print(_snapshot_file)
                return
            logging.info(f"Could not create snapshot at {_snapshot_file} as file is already existing ({i_try + 1}/10)")
        logging.info(f"Could not create any snapshot after 10/10 tries.")

    def load_snapshot(self, file):
        logging.info(f"load snapshot for preprocessing from {file}")
        with open(file, "rb") as f:
            snapshot = dill.load(f)
        excluded_params = ["activation", "activation_output", "add_dense_layer", "batch_normalization", "batch_path",
                           "batch_size", "block_length", "bootstrap_method", "bootstrap_path", "bootstrap_type",
                           "competitor_path", "competitors", "create_new_bootstraps", "create_new_model",
                           "create_snapshot", "data_collection", "debug_mode", "dense_layer_configuration",
                           "do_uncertainty_estimate", "dropout", "dropout_rnn", "early_stopping_epochs", "epochs",
                           "evaluate_competitors", "evaluate_feature_importance", "experiment_name", "experiment_path",
                           "exponent_last_layer", "forecast_path", "fraction_of_training", "hostname", "hpc_hosts",
                           "kernel_regularizer", "kernel_size", "layer_configuration", "log_level_stream",
                           "logging_path", "login_nodes", "loss_type", "loss_weights", "max_number_multiprocessing",
                           "model_class", "model_display_name", "model_path", "n_boots", "n_hidden", "n_layer",
                           "neighbors", "plot_list", "plot_path", "regularizer", "restore_best_model_weights",
                           "snapshot_load_path", "snapshot_path", "stations", "tmp_path", "train_model",
                           "transformation", "use_multiprocessing", "cams_data_path", "cams_interp_method", 
                           "do_bias_free_evaluation", "apriori_file", "model_path", "model_load_path", "era5_data_path",
                           "era5_file_names", "ifs_data_path", "ifs_file_names"]
        data_handler = self.data_store.get("data_handler")
        model_class = self.data_store.get("model_class")
        excluded_params = list(set(excluded_params + data_handler.store_attributes() + model_class.requirements()))

        if check_nested_equality(self.data_store._store, snapshot._store, skip_args=excluded_params) is True:
            self.update_datastore(snapshot, excluded_params=remove_items(excluded_params, ["transformation",
                                                                                           "data_collection",
                                                                                           "stations"]))
        else:
            raise ReferenceError("provided snapshot does not match with the current experiment setup. Abort this run!")


def f_proc(data_handler, station, name_affix, store, return_strategy="", tmp_path=None, **kwargs):
    """
    Try to create a data handler for given arguments. If build fails, this station does not fulfil all requirements and
    therefore f_proc will return None as indication. On a successful build, f_proc returns the built data handler and
    the station that was used. This function must be implemented globally to work together with multiprocessing.
    """
    assert return_strategy in ["result", "reference"]
    try:
        res = data_handler.build(station, name_affix=name_affix, store_processed_data=store, **kwargs)
    except (AttributeError, EmptyQueryResult, KeyError, requests.ConnectionError, ValueError, IndexError) as e:
        formatted_lines = traceback.format_exc().splitlines()
        logging.info(f"remove station {station} because it raised an error: {e} -> "
                     f"{' | '.join(f_inspect_error(formatted_lines))}")
        logging.debug(f"detailed information for removal of station {station}: {traceback.format_exc()}")
        res = None
    if return_strategy == "result":
        return res, station
    else:
        if tmp_path is None:
            tmp_path = os.getcwd()
        _tmp_file = os.path.join(tmp_path, f"{station}_{'%032x' % random.getrandbits(128)}.pickle")
        with open(_tmp_file, "wb") as f:
            dill.dump(res, f, protocol=4)
        return _tmp_file, station


def f_proc_create_info_df(data, meta_cols):
    station_name = str(data.id_class)
    meta = data.id_class.meta
    res = {"station_name": station_name, "Y_shape": data.get_Y()[0].shape[0],
           "meta": meta.reindex(meta_cols).values.flatten()}
    return res


def f_inspect_error(formatted):
    for i in range(len(formatted) - 1, -1, -1):
        if "mlair/mlair" not in formatted[i]:
            return formatted[i - 3:i]
    return formatted[-3:0]
